from __future__ import annotations

from pathlib import Path
from typing import ClassVar

from attrs import frozen

from ...settings._patroni import Settings
from .. import impl


@frozen
class Service:
    """A Patroni service bound to a PostgreSQL instance."""

    __service_name__: ClassVar = "patroni"
    cluster: str
    node: str
    name: str
    settings: Settings

    def __str__(self) -> str:
        return f"{self.__service_name__}@{self.name}"

    def args(self) -> list[str]:
        configpath = impl._configpath(self.name, self.settings)
        return [str(self.settings.execpath), str(configpath)]

    def pidfile(self) -> Path:
        return Path(str(self.settings.pid_file).format(name=self.name))

    def env(self) -> None:
        return None



import os
import shutil
from pathlib import Path
from typing import Any, ClassVar

from pydantic import Field, root_validator, validator

from .. import util
from .base import BaseModel


def default_systemd_unit_path(uid: int) -> Path:
    """Return the default systemd unit path for 'uid'.

    >>> default_systemd_unit_path(0)
    PosixPath('/etc/systemd/system')
    >>> default_systemd_unit_path(42)  # doctest: +ELLIPSIS
    PosixPath('/.../.local/share/systemd/user')
    """
    if uid == 0:
        return Path("/etc/systemd/system")
    return util.xdg_data_home() / "systemd" / "user"


class Settings(BaseModel):
    """Systemd settings."""

    systemctl: ClassVar[Path]

    @root_validator
    def __systemctl_(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not hasattr(cls, "systemctl"):
            systemctl = shutil.which("systemctl")
            if systemctl is None:
                raise ValueError("systemctl command not found")
            cls.systemctl = Path(systemctl)  # type: ignore[misc]
        return values

    unit_path: Path = Field(
        default=default_systemd_unit_path(os.getuid()),
        description="Base path where systemd units will be installed.",
    )

    user: bool = Field(
        default=True,
        description="Use the system manager of the calling user, by passing --user to systemctl calls.",
    )

    sudo: bool = Field(
        default=False,
        description="Run systemctl command with sudo; only applicable when 'user' is unset.",
    )

    @validator("sudo")
    def __validate_sudo_and_user_(cls, value: bool, values: dict[str, Any]) -> bool:
        if value and values.get("user"):
            raise ValueError("cannot be used with 'user' mode")
        return value



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Literal, Optional

import click
import pluggy
from pgtoolkit.conf import Configuration
from pydantic.fields import FieldInfo
from typing_extensions import TypeAlias

from . import __name__ as pkgname

if TYPE_CHECKING:
    from .ctx import Context
    from .models import interface, system
    from .models.system import BaseInstance, Instance, PostgreSQLInstance
    from .pm import PluginManager
    from .postgresql import Standby
    from .settings import Settings
    from .types import ConfigChanges, ServiceManifest, Status

hookspec = pluggy.HookspecMarker(pkgname)

FirstResult: TypeAlias = Optional[Literal[True]]


@hookspec
def site_configure_install(
    settings: Settings, pm: PluginManager, header: str, env: str | None
) -> None:
    """Global configuration hook."""
    raise NotImplementedError


@hookspec
def site_configure_uninstall(
    settings: Settings, pm: PluginManager, ctx: Context
) -> None:
    """Global configuration hook."""
    raise NotImplementedError


@hookspec
def systemd_unit_templates(
    settings: Settings, env: dict[str, Any], content: bool
) -> Iterator[str] | Iterator[tuple[str, str]]:
    """Systemd unit templates used by each plugin."""
    raise NotImplementedError


@hookspec
def site_configure_installed(settings: Settings, pm: PluginManager) -> bool:
    """Check installation in each plugin, returning True is installation is okay."""
    raise NotImplementedError


@hookspec
def cli() -> click.Command:
    """Return command-line entry point as click Command (or Group) for the plugin."""
    raise NotImplementedError


@hookspec
def instance_cli(group: click.Group) -> None:
    """Extend 'group' with extra commands from the plugin."""
    raise NotImplementedError


@hookspec
def system_lookup(instance: PostgreSQLInstance) -> Any | None:
    """Look up for the satellite service object on system that matches specified instance."""
    raise NotImplementedError


@hookspec
def get(instance: Instance) -> ServiceManifest | None:
    """Return the description the satellite service bound to specified instance."""
    raise NotImplementedError


@hookspec
def interface_model() -> tuple[str, Any, Any]:
    """The interface model for satellite component provided plugin."""
    raise NotImplementedError


@hookspec
def instance_settings(
    manifest: interface.Instance, instance: BaseInstance
) -> Configuration:
    """Called before the PostgreSQL instance settings is written."""
    raise NotImplementedError


@hookspec(firstresult=True)
def standby_model(
    ctx: Context,
    instance: system.Instance,
    standby: system.Standby,
    running: bool,
) -> Standby | None:
    """The interface model holding standby information, if 'instance' is a
    plain standby.

    Only one implementation should be invoked so call order and returned value
    matter.

    An implementation may raise a ValueError to interrupt hook execution.
    """
    raise NotImplementedError


@hookspec
def instance_configured(
    ctx: Context,
    manifest: interface.Instance,
    config: Configuration,
    changes: ConfigChanges,
    creating: bool,
    upgrading_from: Instance | None,
) -> None:
    """Called when the PostgreSQL instance got (re-)configured."""
    raise NotImplementedError


@hookspec
def instance_dropped(ctx: Context, instance: Instance) -> None:
    """Called when the PostgreSQL instance got dropped."""
    raise NotImplementedError


@hookspec
def instance_started(ctx: Context, instance: Instance) -> None:
    """Called when the PostgreSQL instance got started."""
    raise NotImplementedError


@hookspec
def instance_stopped(ctx: Context, instance: Instance) -> None:
    """Called when the PostgreSQL instance got stopped."""
    raise NotImplementedError


@hookspec
def instance_promoted(instance: Instance) -> None:
    """Called when the PostgreSQL instance got promoted."""
    raise NotImplementedError


@hookspec
def instance_env(instance: Instance) -> dict[str, str]:
    """Return environment variables for instance defined by the plugin."""
    raise NotImplementedError


@hookspec
def instance_upgraded(old: PostgreSQLInstance, new: PostgreSQLInstance) -> None:
    """Called when 'old' PostgreSQL instance got upgraded as 'new'."""
    raise NotImplementedError


@hookspec
def instance_status(ctx: Context, instance: Instance) -> tuple[Status, str] | None:
    """Return instance status"""
    raise NotImplementedError


@hookspec
def role_model() -> tuple[str, Any, FieldInfo]:
    """Return the definition for an extra field to the Role interface model
    provided by a plugin.
    """
    raise NotImplementedError


@hookspec
def role_change(role: interface.BaseRole, instance: PostgreSQLInstance) -> bool:
    """Called when 'role' changed in 'instance' (be it a create, an update or a deletion).

    Return True if any change happened during hook invocation.
    """
    raise NotImplementedError


@hookspec
def role_inspect(instance: PostgreSQLInstance, name: str) -> dict[str, Any]:
    """Return extra attributes for 'name' role from plugins."""
    raise NotImplementedError


@hookspec
def rolename(settings: Settings) -> str:
    """Return the name of role used by a plugin."""
    raise NotImplementedError


@hookspec
def role(settings: Settings, manifest: interface.Instance) -> interface.Role:
    """Return the role used by a plugin, to be created at instance creation."""
    raise NotImplementedError


@hookspec
def database(settings: Settings, manifest: interface.Instance) -> interface.Database:
    """Return the database used by a plugin, to be created at instance creation."""
    raise NotImplementedError


@hookspec(firstresult=True)
def initdb(
    ctx: Context, manifest: interface.Instance, instance: BaseInstance
) -> FirstResult:
    """Initialize a PostgreSQL database cluster.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec
def patroni_create_replica_method(
    manifest: interface.Instance,
    instance: BaseInstance,
) -> tuple[str, dict[str, Any]] | None:
    raise NotImplementedError


@hookspec(firstresult=True)
def postgresql_editable_conf(ctx: Context, instance: BaseInstance) -> str | None:
    """Return the content of editable postgresql.conf.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def configure_postgresql(
    ctx: Context,
    manifest: interface.Instance,
    configuration: Configuration,
    instance: BaseInstance,
) -> ConfigChanges | None:
    """Configure PostgreSQL and return 'changes' to postgresql.conf.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def configure_auth(
    settings: Settings, instance: BaseInstance, manifest: interface.Instance
) -> bool | None:
    """Configure authentication for PostgreSQL (pg_hba.conf, pg_ident.conf).

    Only one implementation should be invoked so call order and returned value
    matter.

    If returning True, PostgreSQL should be restarted by the caller.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def start_postgresql(
    ctx: Context, instance: PostgreSQLInstance, foreground: bool, wait: bool
) -> FirstResult:
    """Start PostgreSQL for specified 'instance'.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def stop_postgresql(
    ctx: Context,
    instance: PostgreSQLInstance,
    mode: str,
    wait: bool,
    deleting: bool,
) -> FirstResult:
    """Stop PostgreSQL for specified 'instance'.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def restart_postgresql(
    ctx: Context, instance: Instance, mode: str, wait: bool
) -> FirstResult:
    """Restart PostgreSQL for specified 'instance'.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def reload_postgresql(ctx: Context, instance: Instance) -> FirstResult:
    """Reload PostgreSQL configuration for 'instance'.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def promote_postgresql(ctx: Context, instance: Instance) -> FirstResult:
    """Promote PostgreSQL for 'instance'.

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def postgresql_service_name(ctx: Context, instance: BaseInstance) -> str:
    """Return the system service name (e.g.  postgresql).

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def enable_service(ctx: Context, service: str, name: str | None) -> FirstResult:
    """Enable a service

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def disable_service(
    ctx: Context, service: str, name: str | None, now: bool | None
) -> FirstResult:
    """Disable a service

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def start_service(ctx: Context, service: str, name: str | None) -> FirstResult:
    """Start a service for a plugin

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def stop_service(ctx: Context, service: str, name: str | None) -> FirstResult:
    """Stop a service for a plugin

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def restart_service(ctx: Context, service: str, name: str | None) -> FirstResult:
    """Restart a service for a plugin

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def service_status(ctx: Context, service: str, name: str | None) -> Status:
    """Return a service status for a plugin

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def schedule_service(ctx: Context, service: str, name: str) -> FirstResult:
    """Schedule a job through timer

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def unschedule_service(
    ctx: Context, service: str, name: str, now: bool | None
) -> FirstResult:
    """Unchedule a job

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def start_timer(ctx: Context, service: str, name: str) -> FirstResult:
    """Start a timer

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def stop_timer(ctx: Context, service: str, name: str) -> FirstResult:
    """Stop a timer

    Only one implementation should be invoked so call order and returned value
    matter.
    """
    raise NotImplementedError


@hookspec
def logrotate_config(settings: Settings) -> str:
    """Return logrotate configuration for the service matching specified instance."""
    raise NotImplementedError



from __future__ import annotations

import logging
from pathlib import Path

from pglift import util

from .. import hookimpl
from ..models.system import Instance, PostgreSQLInstance
from ..pm import PluginManager
from ..settings import Settings, _logrotate

logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    return settings.logrotate is not None


def get_settings(settings: Settings) -> _logrotate.Settings:
    assert settings.logrotate is not None
    return settings.logrotate


def config_path(settings: _logrotate.Settings) -> Path:
    return settings.configdir / "logrotate.conf"


@hookimpl
def site_configure_install(settings: Settings, pm: PluginManager) -> None:
    logger.info("creating logrotate config directory")
    s = get_settings(settings)
    s.configdir.mkdir(mode=0o750, exist_ok=True, parents=True)
    results = pm.hook.logrotate_config(settings=settings)
    with config_path(s).open("w") as f:
        logger.info("writing logrotate config")
        f.write("\n".join(results))


@hookimpl
def site_configure_uninstall(settings: Settings) -> None:
    logger.info("deleting logrotate config directory")
    s = get_settings(settings)
    util.rmtree(s.configdir)


@hookimpl
def site_configure_installed(settings: Settings) -> bool:
    s = get_settings(settings)
    if not (fpath := config_path(s)).exists():
        logger.error("logrotate configuration '%s' missing", fpath)
        return False
    return True


def instance_configpath(
    settings: _logrotate.Settings, instance: PostgreSQLInstance
) -> Path:
    return settings.configdir / f"{instance.qualname}.conf"


@hookimpl
def instance_dropped(instance: Instance) -> None:
    settings = get_settings(instance._settings)
    instance_configpath(settings, instance).unlink(missing_ok=True)



from __future__ import annotations

import pathlib

import httpx
import pytest
from pgtoolkit.conf import Configuration

import pglift.pgbackrest.base as pgbackrest
import pglift.pgbackrest.models as pgbackrest_models
import pglift.prometheus.impl as prometheus
import pglift.prometheus.models as prometheus_models
from pglift import exceptions, systemd
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.settings import Settings, _pgbackrest
from pglift.systemd import scheduler, service_manager

from .pgbackrest import PgbackrestRepoHost


@pytest.mark.standby
def test_standby_pgbackrest_teardown(
    instance: system.Instance,
    standby_instance: system.Instance,
    standby_instance_dropped: Configuration,
) -> None:
    pgbackrest_settings = pgbackrest.available(standby_instance._settings)
    if not pgbackrest_settings:
        pytest.skip("pgbackrest not available")
    assert not pgbackrest.enabled(standby_instance, pgbackrest_settings)
    assert (
        pgbackrest.system_lookup(standby_instance.datadir, pgbackrest_settings) is None
    )
    standby_svc = standby_instance.service(pgbackrest_models.Service)  # still in memory
    config = standby_svc.path.read_text()
    assert f"pg{standby_svc.index}-path" not in config
    assert str(standby_instance.datadir) not in config

    svc = instance.service(pgbackrest_models.Service)
    assert svc.path == standby_svc.path
    assert f"pg{svc.index}-path" in config
    assert str(instance.datadir) in config


def test_upgrade_pgbackrest_teardown(
    to_be_upgraded_instance: system.Instance,
    upgraded_instance: system.Instance,
    to_be_upgraded_instance_dropped: Configuration,
    upgraded_instance_dropped: Configuration,
) -> None:
    pgbackrest_settings = pgbackrest.available(to_be_upgraded_instance._settings)
    if not pgbackrest_settings:
        pytest.skip("pgbackrest not available")
    assert pgbackrest.available(upgraded_instance._settings) == pgbackrest_settings

    assert not pgbackrest.enabled(to_be_upgraded_instance, pgbackrest_settings)
    assert (
        pgbackrest.system_lookup(to_be_upgraded_instance.datadir, pgbackrest_settings)
        is None
    )
    svc = to_be_upgraded_instance.service(pgbackrest_models.Service)  # still in memory
    assert not svc.path.exists()

    assert not pgbackrest.enabled(upgraded_instance, pgbackrest_settings)
    assert (
        pgbackrest.system_lookup(upgraded_instance.datadir, pgbackrest_settings) is None
    )
    svc = upgraded_instance.service(pgbackrest_models.Service)  # still in memory
    assert not svc.path.exists()


@pytest.mark.standby
def test_pgbackrest_teardown(
    instance: system.Instance,
    standby_instance_dropped: Configuration,
    instance_dropped: Configuration,
    pgbackrest_repo_host: PgbackrestRepoHost | None,
) -> None:
    pgbackrest_settings = pgbackrest.available(instance._settings)
    if not pgbackrest_settings:
        pytest.skip("pgbackrest not available")
    stanza = f"mystanza-{instance.name}"
    path_repository: _pgbackrest.PathRepository | None = None
    if pgbackrest_repo_host is None:
        assert isinstance(pgbackrest_settings.repository, _pgbackrest.PathRepository)
        path_repository = pgbackrest_settings.repository
    if path_repository:
        # Backups are kept (confirmation prompt defaults to 'no', in
        # instance_drop() hook).
        assert list(
            (path_repository.path / "archive").glob(f"{stanza}*"),
        )
        assert list(
            (path_repository.path / "backup").glob(f"{stanza}*"),
        )
    assert not (pgbackrest_settings.configpath / "conf.d" / f"{stanza}.conf").exists()
    # global directories and files are preserved
    assert (pgbackrest_settings.configpath / "pgbackrest.conf").exists()
    assert (pgbackrest_settings.configpath / "conf.d").exists()
    if path_repository:
        assert path_repository.path.exists()
    assert pgbackrest_settings.spoolpath.exists()
    assert pgbackrest_settings.logpath.exists()
    assert not list(pgbackrest_settings.logpath.glob(f"{stanza}*.log"))


def test_pgpass(
    settings: Settings,
    passfile: pathlib.Path,
    instance_manifest: interface.Instance,
    to_be_upgraded_instance_dropped: Configuration,
    upgraded_instance_dropped: Configuration,
    instance_dropped: Configuration,
) -> None:
    assert not passfile.exists()


@pytest.mark.usefixtures("require_systemd_scheduler")
def test_systemd_backup_job(
    ctx: Context, instance: system.Instance, instance_dropped: Configuration
) -> None:
    unit = scheduler.unit("backup", instance.qualname)
    assert not systemd.is_active(ctx, unit)
    assert not systemd.is_enabled(ctx, unit)


def test_prometheus_teardown(
    ctx: Context,
    instance: system.Instance,
    instance_dropped: Configuration,
) -> None:
    prometheus_settings = prometheus.available(ctx.settings)
    if not prometheus_settings:
        pytest.skip("prometheus not available")
    configpath = pathlib.Path(
        str(prometheus_settings.configpath).format(name=instance.qualname)
    )
    queriespath = pathlib.Path(
        str(prometheus_settings.queriespath).format(name=instance.qualname)
    )
    assert not configpath.exists()
    assert not queriespath.exists()
    if ctx.settings.service_manager == "systemd":
        assert not systemd.is_enabled(
            ctx, service_manager.unit("postgres_exporter", instance.qualname)
        )
        service = instance.service(prometheus_models.Service)
        port = service.port
        with pytest.raises(httpx.ConnectError):
            httpx.get(f"http://0.0.0.0:{port}/metrics")


def test_databases_teardown(
    instance: system.Instance,
    instance_dropped: Configuration,
) -> None:
    assert not instance.dumps_directory.exists()


def test_instance(
    ctx: Context, instance: system.Instance, instance_dropped: Configuration
) -> None:
    with pytest.raises(exceptions.InstanceNotFound, match=str(instance)):
        instance.exists()



from __future__ import annotations

import pytest

from pglift import _install
from pglift.pm import PluginManager
from pglift.settings import Settings


def test_check_uninstalled(pm: PluginManager, settings: Settings) -> None:
    assert not _install.check(pm, settings)


@pytest.mark.usefixtures("installed")
def test_check_installed(pm: PluginManager, settings: Settings) -> None:
    assert _install.check(pm, settings)



from __future__ import annotations

import logging
import re
import socket
import subprocess
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from unittest.mock import patch

import httpx
import pgtoolkit.conf
import pytest
import yaml
from tenacity import Retrying, retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed

from pglift import instances, postgresql, systemd
from pglift._compat import zip
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.patroni import impl as patroni
from pglift.patroni.models import Patroni, Service, ServiceManifest
from pglift.pgbackrest import base as pgbackrest
from pglift.pgbackrest import models as pgbackrest_models
from pglift.pgbackrest import repo_path
from pglift.settings import Settings, _patroni
from pglift.systemd import service_manager
from pglift.types import ConfigChanges, Status

from .. import CertFactory
from ..etcd import Etcd
from . import AuthType, PostgresLogger, check_connect, passfile_entries
from .conftest import Factory
from .pgbackrest import PgbackrestRepoHost


@contextmanager
def reconfigure_instance(
    ctx: Context,
    manifest: interface.Instance,
    **confitems: Any,
) -> Iterator[ConfigChanges]:
    """Context manager to temporarily change instance settings.

    Upon enter, this applies provided settings (and possibly new port)
    and yields settings 'changes' dict.

    Upon exit, the previous settings is restored, and the 'changes' dict
    returned upon enter is updated to reflect this.
    """
    update: dict[str, Any] = {"restart_on_changes": manifest.restart_on_changes}
    if confitems:
        update["settings"] = manifest.settings.copy() | confitems
    assert update
    changes = instances.configure(ctx, manifest._copy_validate(update))
    try:
        yield changes
    finally:
        restored_changes = instances.configure(ctx, manifest)
        changes.clear()
        changes.update(restored_changes)


@pytest.fixture(scope="session", autouse=True)
def _patroni_available(patroni_execpath: Path | None) -> None:
    if not patroni_execpath:
        pytest.skip("Patroni is not available")


@pytest.fixture(scope="module")
def etcd_credentials() -> tuple[str, str]:
    return "patroni", "p@tr0n!"


@pytest.fixture(scope="module", autouse=True)
def _etcd_running(
    etcd_host: Etcd, etcd_credentials: tuple[str, str], cluster_name: str
) -> Iterator[None]:
    with etcd_host.running() as e:
        e.setup_auth(
            credentials=etcd_credentials, role="svc", prefix=f"/service/{cluster_name}"
        )
        yield None


@pytest.fixture(scope="module", autouse=True)
def http_logs() -> None:
    logging.getLogger("httpx").setLevel(logging.DEBUG)
    logging.getLogger("httpcore").setLevel(logging.DEBUG)


@pytest.fixture
def patroni_settings(settings: Settings) -> _patroni.Settings:
    return patroni.get_settings(settings)


@pytest.fixture(scope="module")
def cluster_name(postgresql_auth: AuthType) -> str:
    # Since instances are kept running while moving from one postgresql_auth
    # value to another, we need distinct cluster name for each.
    return f"pglift-tests-{postgresql_auth}"


def _make_instance(
    settings: Settings,
    manifest: interface.Instance,
    postgres_logger: PostgresLogger,
    pgbackrest_repo_host: PgbackrestRepoHost | None,
    pgbackrest_password: str | None,
) -> Iterator[system.Instance]:
    ctx = Context(settings=settings)
    assert instances.apply(ctx, manifest)
    instance = system.Instance.system_lookup(
        (manifest.name, manifest.version, settings)
    )
    postgres_logger(instance)
    if settings.pgbackrest:
        if not instance.standby:
            # make a pgbackrest backup so future standby could be initialized from
            # pgbackrest backup
            if pgbackrest_repo_host is not None:
                svc = instance.service(pgbackrest_models.Service)
                pgbackrest_repo_host.add_stanza(svc.stanza, instance)
                pgbackrest.check(
                    instance, svc, settings.pgbackrest, pgbackrest_password
                )
                pgbackrest_repo_host.run(
                    "backup",
                    "--stanza",
                    svc.stanza,
                    "--type",
                    "full",
                )
            else:
                repo_path.backup(instance, settings.pgbackrest)
        else:
            # ensure standby has been created from pgbackrest backup
            assert instance._settings.patroni is not None
            assert any(
                "replica has been created using pgbackrest" in line
                for line in patroni.logs(instance.qualname, instance._settings.patroni)
            )
    yield instance
    if instances.exists(instance.name, instance.version, instance._settings):
        instances.drop(ctx, instance)


@pytest.fixture(scope="module")
def instance1_manifest(
    settings: Settings,
    instance_manifest_factory: Factory[interface.Instance],
    cluster_name: str,
    tmp_port_factory: Iterator[int],
    ca_cert: Path,
    cert_factory: CertFactory,
    etcd_credentials: tuple[str, str],
) -> interface.Instance:
    name = "test1"
    hostname = socket.gethostname()
    host = socket.gethostbyname(hostname)
    extras = {}
    if settings.pgbackrest:
        extras = {"pgbackrest": {"stanza": "patroni"}}
    server_cert = cert_factory(host, common_name=hostname)
    return instance_manifest_factory(
        settings,
        name,
        state="started",
        patroni={
            "cluster": cluster_name,
            "node": name,
            "etcd": {
                "username": etcd_credentials[0],
                "password": etcd_credentials[1],
            },
            "restapi": {"connect_address": f"{host}:{next(tmp_port_factory)}"},
        },
        auth={"host": "password"},
        settings={
            "listen_addresses": "*",
            "work_mem": "8MB",
            "ssl": True,
            "ssl_ca_file": ca_cert,
            "ssl_cert_file": server_cert.path,
            "ssl_key_file": server_cert.private_key,
            "log_connections": True,
        },
        **extras,
    )


@pytest.fixture(scope="module")
def instance1(
    settings: Settings,
    instance1_manifest: interface.Instance,
    postgres_logger: PostgresLogger,
    pgbackrest_repo_host: PgbackrestRepoHost | None,
    pgbackrest_password: str | None,
) -> Iterator[system.Instance]:
    yield from _make_instance(
        settings,
        instance1_manifest,
        postgres_logger,
        pgbackrest_repo_host,
        pgbackrest_password,
    )


@pytest.fixture(scope="module")
def instance2_manifest(
    settings: Settings,
    instance_manifest_factory: Factory[interface.Instance],
    cluster_name: str,
    tmp_port_factory: Iterator[int],
    instance1: system.Instance,
    cert_factory: CertFactory,
    etcd_credentials: tuple[str, str],
) -> interface.Instance:
    name = "test2"
    host = socket.gethostbyname(socket.gethostname())
    extras = {}
    if settings.pgbackrest:
        extras = {"pgbackrest": {"stanza": "patroni"}}
    replication_cert = cert_factory(common_name="replication")
    return instance_manifest_factory(
        settings,
        name,
        state="started",
        patroni={
            "cluster": cluster_name,
            "node": name,
            "postgresql": {
                "replication": {
                    "ssl": {
                        "cert": replication_cert.path,
                        "key": replication_cert.private_key,
                    },
                },
            },
            "etcd": {
                "username": etcd_credentials[0],
                "password": etcd_credentials[1],
            },
            "restapi": {"connect_address": f"{host}:{next(tmp_port_factory)}"},
        },
        auth={"host": "password"},
        settings={
            "listen_addresses": "*",
            "work_mem": "8MB",
        },
        **extras,
    )


@pytest.fixture(scope="module")
def instance2(
    settings: Settings,
    instance2_manifest: interface.Instance,
    postgres_logger: PostgresLogger,
    pgbackrest_repo_host: PgbackrestRepoHost | None,
    pgbackrest_password: str | None,
) -> Iterator[system.Instance]:
    yield from _make_instance(
        settings,
        instance2_manifest,
        postgres_logger,
        pgbackrest_repo_host,
        pgbackrest_password,
    )


@pytest.fixture
def primary_standby(
    instance1: system.Instance, instance2: system.Instance
) -> tuple[system.Instance, system.Instance]:
    if instance1.standby:
        assert not instance2.standby
        return instance2, instance1
    else:
        assert instance2.standby
        return instance1, instance2


@pytest.fixture
def primary(
    primary_standby: tuple[system.Instance, system.Instance]
) -> system.Instance:
    return primary_standby[0]


@pytest.fixture
def standby(
    primary_standby: tuple[system.Instance, system.Instance]
) -> system.Instance:
    return primary_standby[1]


def test_waldir(instance1: system.Instance, instance2: system.Instance) -> None:
    pg_wal1 = instance1.datadir / "pg_wal"
    assert pg_wal1.is_symlink()
    assert pg_wal1.resolve() == instance1.waldir

    pg_wal2 = instance2.datadir / "pg_wal"
    assert pg_wal2.is_symlink()
    assert pg_wal2.resolve() == instance2.waldir


def test_service_and_config(
    patroni_settings: _patroni.Settings,
    instance1: system.Instance,
    instance1_manifest: interface.Instance,
    instance2: system.Instance,
    instance2_manifest: interface.Instance,
    cluster_name: str,
) -> None:
    for instance, manifest in (
        (instance1, instance1_manifest),
        (instance2, instance2_manifest),
    ):
        check_server_and_config(instance, manifest, patroni_settings, cluster_name)


def test_pgpass(
    patroni_settings: _patroni.Settings,
    primary: system.Instance,
    standby: system.Instance,
    replrole_password: str,
) -> None:
    primary_pgpass = patroni._pgpass(primary.qualname, patroni_settings.postgresql)
    assert not primary_pgpass.exists()
    standby_pgpass = patroni._pgpass(standby.qualname, patroni_settings.postgresql)
    assert standby_pgpass.exists()
    (replication_entry,) = passfile_entries(standby_pgpass, role="replication")
    assert replication_entry.endswith(
        f":{primary.port}:*:replication:{replrole_password}"
    )


def check_server_and_config(
    instance: system.Instance,
    manifest: interface.Instance,
    settings: _patroni.Settings,
    cluster_name: str,
) -> None:
    s = instance.service(Service)
    assert s and s.cluster == cluster_name
    configpath = patroni._configpath(instance.qualname, settings)
    with configpath.open() as f:
        config = yaml.safe_load(f)
    listen_addr = manifest.patroni.restapi.listen  # type: ignore[attr-defined]
    assert config["restapi"]["listen"] == listen_addr
    assert config["postgresql"]["listen"] == f"*:{instance.port}"
    assert config["postgresql"]["parameters"]["listen_addresses"] == "*"
    assert config["postgresql"]["parameters"]["work_mem"] == "8MB"
    assert config["bootstrap"]["dcs"]["loop_wait"] == 1
    assert config["ctl"]["certfile"]


def test_postgresql_conf(instance1: system.Instance) -> None:
    pgconf = pgtoolkit.conf.parse(instance1.datadir / "postgresql.conf")
    assert "lc_messages" in pgconf.as_dict()
    assert "lc_monetary" in pgconf.as_dict()


def test_logpath(
    patroni_settings: _patroni.Settings, instance1: system.Instance
) -> None:
    logpath = patroni_settings.logpath / instance1.qualname
    assert logpath.exists()
    assert (logpath / "patroni.log").exists()


def logs(instance: system.Instance, settings: _patroni.Settings) -> list[str]:
    return [
        line.split("INFO: ", 1)[-1].strip()
        for line in patroni.logs(instance.qualname, settings)
    ]


def test_logs(
    patroni_settings: _patroni.Settings,
    instance1: system.Instance,
    instance2: system.Instance,
) -> None:
    logs1 = logs(instance1, patroni_settings)
    logs2 = logs(instance2, patroni_settings)
    leader = instance1.name
    secondary = instance2.name
    assert f"no action. I am ({leader}), the leader with the lock" in logs1
    assert (
        f"no action. I am ({secondary}), a secondary, and following a leader ({leader})"
        in logs2
    )


@pytest.mark.parametrize(
    "setting,expected",
    [
        ("work_mem", "8MB"),
        ("listen_addresses", "*"),
    ],
)
def test_postgresql_config(
    instance1: system.Instance, setting: str, expected: Any
) -> None:
    pgconf = instance1.config()
    assert pgconf[setting] == expected


def test_configure_postgresql(
    ctx: Context,
    patroni_settings: _patroni.Settings,
    instance1_manifest: interface.Instance,
    instance1: system.Instance,
    tmp_port_factory: Iterator[int],
) -> None:
    postgresql_conf = instance1.datadir / "postgresql.conf"
    mtime = postgresql_conf.stat().st_mtime

    # Retry assertions on postgresql.conf, waiting for patroni reload (1s, per
    # loop_wait).
    @retry(
        retry=(
            retry_if_exception_type(ValueError)
            | retry_if_exception_type(AttributeError)
        ),
        wait=wait_fixed(0.5),
        stop=stop_after_attempt(5),
        reraise=True,
    )
    def check_postgresql_config(work_mem: str, mtime: float) -> float:
        mtime_after = postgresql_conf.stat().st_mtime
        if mtime_after <= mtime:
            raise ValueError
        pgconf = pgtoolkit.conf.parse(postgresql_conf)
        assert pgconf.work_mem == work_mem
        return mtime_after

    with reconfigure_instance(ctx, instance1_manifest, work_mem="10MB") as changes:
        config = Patroni.get(instance1.qualname, patroni_settings)
        assert config.postgresql.parameters["work_mem"] == "10MB"
        mtime = check_postgresql_config("10MB", mtime)
        assert changes == {"work_mem": ("8MB", "10MB")}

    assert changes == {"work_mem": ("10MB", "8MB")}
    config = Patroni.get(instance1.qualname, patroni_settings)
    assert config.postgresql.parameters["work_mem"] == "8MB"
    check_postgresql_config("8MB", mtime)


def test_instance_get(
    ctx: Context,
    patroni_settings: _patroni.Settings,
    instance1: system.Instance,
    instance2: system.Instance,
    surole_password: str | None,
    cluster_name: str,
) -> None:
    for instance in (instance1, instance2):
        if surole_password is not None and instance is instance2:
            with patch.dict("os.environ", {"PGPASSWORD": surole_password}):
                i = instances.get(ctx, instance)
        else:
            i = instances.get(ctx, instance)
        p = i.service_manifest(ServiceManifest)
        assert p is not None and p.cluster == cluster_name
        assert p.cluster == cluster_name
        assert {m.name for m in p.cluster_members} == {"test1", "test2"}


def test_check_api_status(
    patroni_settings: _patroni.Settings, instance1: system.Instance
) -> None:
    assert patroni.check_api_status(instance1.qualname, patroni_settings)


def test_cluster_members(
    patroni_settings: _patroni.Settings,
    instance1: system.Instance,
    instance2: system.Instance,
) -> None:
    members = patroni.cluster_members(instance1.qualname, patroni_settings)
    assert len(members) == 2, members
    for m, i in zip(members, (instance1, instance2), strict=True):
        assert m.port == i.port


def test_cluster_leader(
    patroni_settings: _patroni.Settings,
    instance1: system.Instance,
    instance2: system.Instance,
) -> None:
    assert patroni.cluster_leader(instance1.qualname, patroni_settings) == "test1"
    assert patroni.cluster_leader(instance2.qualname, patroni_settings) == "test1"


def test_replication_connection_uses_ssl_cert(
    instance1: system.Instance, instance2: system.Instance
) -> None:
    patterns = [
        r'connection authenticated: identity="CN=replication,.+" method=cert',
        r"replication connection authorized: user=replication application_name=test2 SSL enabled",
    ]
    try:
        for line in postgresql.logs(instance1, timeout=0):
            p = patterns[0]
            if re.search(p, line.rstrip()):
                del patterns[0]
                if not patterns:
                    break
        else:
            pytest.fail(f"expected log lines not found: {patterns!r}")
    except TimeoutError:
        pass


def test_connect(
    ctx: Context,
    postgresql_auth: AuthType,
    instance1_manifest: interface.Instance,
    instance1: system.Instance,
    instance2_manifest: interface.Instance,
    instance2: system.Instance,
) -> None:
    check_connect(ctx, postgresql_auth, instance1_manifest, instance1)
    check_connect(ctx, postgresql_auth, instance2_manifest, instance2)


def test_reload(ctx: Context, instance1: system.Instance) -> None:
    instances.reload(ctx, instance1)


def test_start_restart_stop(
    ctx: Context,
    settings: Settings,
    instance1: system.Instance,
    instance2: system.Instance,
    patroni_settings: _patroni.Settings,
) -> None:
    use_systemd = settings.service_manager == "systemd"

    assert postgresql.is_running(instance1)
    if use_systemd:
        assert systemd.is_active(
            ctx, service_manager.unit("patroni", instance1.qualname)
        )
    assert patroni.check_api_status(instance1.qualname, patroni_settings)

    # Stop instance2, then restart instance1, so that the latter remains
    # leader.
    with instances.stopped(ctx, instance2):
        assert postgresql.status(instance2) == Status.not_running
        if use_systemd:
            assert not systemd.is_active(
                ctx, service_manager.unit("patroni", instance2.qualname)
            )
        assert not patroni.check_api_status(instance2.qualname, patroni_settings)
        instances.restart(ctx, instance1)
    assert postgresql.is_running(instance1)
    assert patroni.check_api_status(instance1.qualname, patroni_settings)

    # Starting instance2 can take a bit of time, so use a retry logic.
    config2 = Patroni.get(instance2.qualname, patroni_settings)
    for attempt in Retrying(
        retry=retry_if_exception_type(httpx.HTTPError),
        wait=wait_fixed(0.5),
        stop=stop_after_attempt(5),
    ):
        with attempt:
            patroni.api_request(config2, "GET", "readiness")


def test_ctl_version(
    primary: system.Instance,
    patroni_settings: _patroni.Settings,
    standby: system.Instance,
    cluster_name: str,
) -> None:
    # Test patronictl with command that uses a REST API endpoint
    configfile = patroni._configpath(primary.qualname, patroni_settings)
    if not (patronictl := patroni_settings.execpath.parent / "patronictl").exists():
        pytest.skip("patronictl executable not found")
    r = subprocess.run(
        [patronictl, "--config-file", configfile, "version", cluster_name],
        check=True,
        capture_output=True,
        text=True,
    )
    patterns = [
        r"patronictl version .*",
        r"test1: Patroni .* PostgreSQL .*$",
        r"test2: Patroni .* PostgreSQL .*$",
    ]
    for line in r.stdout.splitlines():
        p = patterns[0]
        if re.search(p, line.rstrip()):
            del patterns[0]
            if not patterns:
                break
    else:
        pytest.fail(f"expected lines not found: {patterns!r}")



from pydantic import Field

from .base import BaseModel, ConfigPath


class Settings(BaseModel):
    """Settings for rsyslog."""

    configdir: ConfigPath = Field(
        default=ConfigPath("rsyslog"), description="rsyslog config directory"
    )



import string
from collections.abc import Callable, Iterator
from pathlib import Path, PosixPath
from typing import Any, Union

import pydantic
from pydantic.fields import ModelField

from .._compat import Self


def string_format_variables(fmt: str) -> set[str]:
    return {v for _, v, _, _ in string.Formatter().parse(fmt) if v is not None}


class BaseModel(pydantic.BaseModel):
    class Config:
        frozen = True
        extra = pydantic.Extra.forbid
        smart_union = True


class PrefixedPath(PosixPath):
    basedir = Path("")
    key = "prefix"

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable[..., "PrefixedPath"]]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Path, field: ModelField) -> Self:
        if not isinstance(value, cls):
            value = cls(value)
        # Ensure all template variables used in default field value are also
        # used in user value and that no unhandled variables are used.
        expected = string_format_variables(str(field.default))
        if expected != string_format_variables(str(value)):
            raise ValueError(
                "value contains unknown or missing template variable(s); "
                f"expecting: {', '.join(sorted(expected)) or 'none'}"
            )
        return value

    def prefix(self, prefix: Union[str, Path]) -> Path:
        """Return the path prefixed if is not yet absolute.

        >>> PrefixedPath("documents").prefix("/home/alice")
        PosixPath('/home/alice/documents')
        >>> PrefixedPath("/root").prefix("/whatever")
        PosixPath('/root')
        """
        if self.is_absolute():
            return Path(self)
        assert Path(prefix).is_absolute(), (
            f"expecting an absolute prefix (got {prefix!r})",
        )
        return prefix / self.basedir / self


class ConfigPath(PrefixedPath):
    basedir = Path("etc")


class RunPath(PrefixedPath):
    basedir = Path("")
    key = "run_prefix"


class DataPath(PrefixedPath):
    basedir = Path("srv")


class LogPath(PrefixedPath):
    basedir = Path("log")


def prefix_values(values: dict[str, Any], prefixes: dict[str, Path]) -> dict[str, Any]:
    for key, child in values.items():
        if isinstance(child, PrefixedPath):
            values[key] = child.prefix(prefixes[child.key])
        elif isinstance(child, pydantic.BaseModel):
            child_values = {k: getattr(child, k) for k in child.__fields__}
            child_values = prefix_values(child_values, prefixes)
            # Use .construct() to avoid re-validating child.
            values[key] = child.construct(
                _fields_set=child.__fields_set__, **child_values
            )
    return values


class ServerCert(BaseModel):
    """TLS certificate files for a server."""

    ca_cert: pydantic.FilePath = pydantic.Field(
        description="Certificate Authority certificate to verify client requests."
    )
    cert: pydantic.FilePath = pydantic.Field(
        description="Certificate file for TLS encryption."
    )
    key: pydantic.FilePath = pydantic.Field(
        description="Private key for the certificate."
    )



from __future__ import annotations

import logging
import re
import socket
from pathlib import Path
from typing import NoReturn

import psycopg
import pytest
from pgtoolkit import conf as pgconf

from pglift import databases, exceptions, instances, postgresql, systemd
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.settings import Settings
from pglift.systemd import service_manager
from pglift.types import Status

from .. import CertFactory
from . import (
    AuthType,
    check_connect,
    connect,
    execute,
    passfile_entries,
    running_instance,
)
from .conftest import Factory


def test_directories(instance: system.Instance) -> None:
    assert instance.datadir.exists()
    assert instance.waldir.exists()
    assert (instance.waldir / "archive_status").is_dir()


def test_config(
    instance: system.Instance, instance_manifest: interface.Instance
) -> None:
    postgresql_conf = instance.datadir / "postgresql.conf"
    assert postgresql_conf.exists()
    pgconfig = pgconf.parse(postgresql_conf)
    assert {k for k, v in pgconfig.entries.items() if not v.commented} & set(
        instance_manifest.settings
    )


def test_psqlrc(instance: system.Instance) -> None:
    assert instance.psqlrc.read_text().strip().splitlines() == [
        f"\\set PROMPT1 '[{instance}] %n@%~%R%x%# '",
        "\\set PROMPT2 ' %R%x%# '",
    ]


def test_systemd(ctx: Context, instance: system.Instance) -> None:
    if ctx.settings.service_manager == "systemd":
        assert systemd.is_enabled(
            ctx, service_manager.unit("postgresql", instance.qualname)
        )


def test_reinit(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Instance already exists, no-op.
    with monkeypatch.context() as m:

        def fail() -> NoReturn:
            raise AssertionError("unexpected called")

        m.setattr(postgresql, "pg_ctl", fail)
        instances.init(
            ctx, interface.Instance(name=instance.name, version=instance.version)
        )


def test_pgpass(
    ctx: Context,
    passfile: Path,
    instance_manifest: interface.Instance,
    instance: system.Instance,
    surole_password: str,
    pgbackrest_password: str,
    pgbackrest_available: bool,
) -> None:
    port = instance.port
    backuprole = ctx.settings.postgresql.backuprole.name

    assert passfile_entries(passfile) == [f"*:{port}:*:postgres:{surole_password}"]
    if pgbackrest_available:
        assert passfile_entries(passfile, role=backuprole) == [
            f"*:{port}:*:{backuprole}:{pgbackrest_password}"
        ]


def test_connect(
    ctx: Context,
    instance_manifest: interface.Instance,
    instance: system.Instance,
    postgresql_auth: AuthType,
) -> None:
    check_connect(ctx, postgresql_auth, instance_manifest, instance)


def test_replrole(instance: system.Instance) -> None:
    (row,) = execute(
        instance,
        # Simplified version of \du in psql.
        "SELECT rolsuper, rolcanlogin, rolreplication,"
        " ARRAY(SELECT b.rolname"
        "       FROM pg_catalog.pg_auth_members m"
        "       JOIN pg_catalog.pg_roles b ON (m.roleid = b.oid)"
        "       WHERE m.member = r.oid) as memberof"
        " FROM pg_catalog.pg_roles r"
        " WHERE rolname = 'replication'",
    )
    assert row == {
        "rolsuper": False,
        "rolcanlogin": True,
        "rolreplication": True,
        "memberof": ["pg_read_all_stats"],
    }


def test_hba(
    ctx: Context,
    instance_manifest: interface.Instance,
    instance: system.Instance,
    postgresql_auth: AuthType,
) -> None:
    hba_path = instance.datadir / "pg_hba.conf"
    hba = hba_path.read_text().splitlines()
    auth_settings = ctx.settings.postgresql.auth
    auth_instance = instance_manifest.auth
    assert auth_instance is not None
    if postgresql_auth == AuthType.peer:
        assert "peer" in hba[0]
    assert (
        f"local   all             all                                     {auth_settings.local}"
        in hba
    )
    assert (
        f"host    all             all             127.0.0.1/32            {auth_instance.host}"
        in hba
    )
    assert (
        f"hostssl all             all             127.0.0.1/32            {auth_settings.hostssl}"
        in hba
    )


def test_ident(
    ctx: Context, instance: system.Instance, postgresql_auth: AuthType
) -> None:
    ident_path = instance.datadir / "pg_ident.conf"
    ident = ident_path.read_text().splitlines()
    assert ident[0] == "# MAPNAME       SYSTEM-USERNAME         PG-USERNAME"
    if postgresql_auth == AuthType.peer:
        assert re.match(r"^test\s+\w+\s+postgres$", ident[1])
    else:
        assert len(ident) == 1


def test_start_stop_restart_running_is_ready_stopped(
    ctx: Context, instance: system.Instance, caplog: pytest.LogCaptureFixture
) -> None:
    i = instance
    assert postgresql.status(i) == Status.running
    assert postgresql.is_ready(i)
    use_systemd = ctx.settings.service_manager == "systemd"
    if use_systemd:
        assert systemd.is_active(ctx, service_manager.unit("postgresql", i.qualname))

    instances.stop(ctx, i)
    try:
        assert postgresql.status(i) == Status.not_running
        assert not postgresql.is_ready(i)
        if use_systemd:
            assert not systemd.is_active(
                ctx, service_manager.unit("postgresql", i.qualname)
            )
        # Stopping a non-running instance is a no-op.
        caplog.clear()
        with caplog.at_level(logging.WARNING, logger="pglift"):
            instances.stop(ctx, i)
        assert f"instance {instance} is already stopped" in caplog.records[0].message
    finally:
        instances.start(ctx, i)

    assert postgresql.status(i) == Status.running
    assert postgresql.is_ready(i)
    if use_systemd:
        assert systemd.is_active(ctx, service_manager.unit("postgresql", i.qualname))

    assert postgresql.status(i) == Status.running
    assert postgresql.is_ready(i)
    if not use_systemd:
        # FIXME: systemctl restart would fail with:
        #   Start request repeated too quickly.
        #   Failed with result 'start-limit-hit'.
        instances.restart(ctx, i)
        assert postgresql.status(i) == Status.running
        assert postgresql.is_ready(i)
    instances.reload(ctx, i)
    assert postgresql.status(i) == Status.running
    assert postgresql.is_ready(i)

    assert postgresql.status(i) == Status.running
    with instances.stopped(ctx, i):
        assert postgresql.status(i) == Status.not_running
        with instances.stopped(ctx, i):
            assert postgresql.status(i) == Status.not_running
            assert not postgresql.is_ready(i)
        with running_instance(ctx, i):
            assert postgresql.status(i) == Status.running
            assert postgresql.is_ready(i)
            with running_instance(ctx, i):
                assert postgresql.status(i) == Status.running
            with instances.stopped(ctx, i):
                assert postgresql.status(i) == Status.not_running
                assert not postgresql.is_ready(i)
            assert postgresql.status(i) == Status.running
            assert postgresql.is_ready(i)
        assert postgresql.status(i) == Status.not_running
    assert postgresql.status(i) == Status.running
    assert postgresql.is_ready(i)


def test_apply(
    ctx: Context,
    pg_version: str,
    tmp_path: Path,
    instance_factory: Factory[tuple[interface.Instance, system.Instance]],
    caplog: pytest.LogCaptureFixture,
    ca_cert: Path,
    cert_factory: CertFactory,
) -> None:
    assert not system.BaseInstance.get("test_apply", pg_version, ctx.settings).exists()
    hostname = socket.gethostname()
    host = socket.gethostbyname(hostname)
    server_cert = cert_factory(host, common_name=hostname)
    im, i = instance_factory(
        ctx.settings,
        name="test_apply",
        settings={
            "unix_socket_directories": str(tmp_path),
            "ssl": True,
            "ssl_cert_file": str(server_cert.path),
            "ssl_key_file": str(server_cert.private_key),
            "ssl_ca_file": str(ca_cert),
            "log_connections": True,
        },
        auth={"hostssl": "cert"},
        restart_on_changes=False,
        roles=[{"name": "bob", "password": "loooooooooong p3w2d!", "login": True}],
        databases=[
            {"name": "db1", "schemas": [{"name": "sales"}]},
            {
                "name": "db2",
                "owner": "bob",
                "extensions": [{"name": "unaccent", "schema": "public"}],
            },
        ],
        pgbackrest={"stanza": "test_apply_stanza"},
    )
    assert i.port == im.port
    pgconfig = i.config()
    assert pgconfig

    # Re-applying the same manifest is a no-op.
    result_apply = instances.apply(ctx, im)
    assert result_apply.change_state is None

    # Change the runtime state.
    assert postgresql.status(i) == Status.not_running
    im = im._copy_validate({"state": "started"})
    result_apply = instances.apply(ctx, im)
    assert result_apply.change_state == interface.ApplyChangeState.changed
    assert postgresql.status(i) == Status.running
    with connect(i) as conn:
        assert not instances.pending_restart(conn)

    with postgresql.running(ctx, i):
        assert databases.exists(ctx, i, "db1")
        db1 = databases.get(ctx, i, "db1")
        assert db1.schemas[1].name == "sales"
        assert databases.exists(ctx, i, "db2")
        db2 = databases.get(ctx, i, "db2")
        assert db2.extensions[0].name == "unaccent"
        assert db2.owner == "bob"

    # Check SSL client connections
    client_cert = cert_factory("127.0.0.1", common_name="bob")
    connargs = {
        "hostaddr": "127.0.0.1",
        "port": i.port,
        "user": "bob",
        "password": "!!! invalid but unused !!!",
        "dbname": "db2",
        "sslmode": "require",
        "sslrootcert": str(ca_cert),
        "application_name": "pglift-tests-ssl",
    }
    with pytest.raises(
        psycopg.OperationalError, match="connection requires a valid client certificate"
    ):
        psycopg.connect(**connargs)  # type: ignore[call-overload]

    connargs |= {
        "sslmode": "verify-ca",
        "sslcert": str(client_cert.path),
        "sslkey": str(client_cert.private_key),
    }

    expected = [
        "connection authorized: user=bob database=db2 application_name=pglift-tests-ssl SSL enabled",
    ]
    if pg_version >= "14":
        expected.insert(0, 'connection authenticated: identity="CN=bob,OU=Testing cert')
    with psycopg.connect(**connargs) as conn:  # type: ignore[call-overload]
        cur = conn.execute("SHOW ssl")
        assert cur.fetchone() == ("on",)
        logs = []
        try:
            for line in postgresql.logs(i, timeout=0):
                logs.append(line.rstrip())
        except TimeoutError:
            pass
        for line in logs:
            if expected[0] in line:
                del expected[0]
                continue
            if not expected:
                break
        if expected:
            pytest.fail(f"expected log line(s) {expected} not found in {logs}")

    # Change PostgreSQL settings.
    newconfig = im.settings.copy()
    newconfig["listen_addresses"] = "*"  # requires restart
    newconfig["autovacuum"] = False  # requires reload
    im = im._copy_validate({"settings": newconfig})
    with caplog.at_level(logging.DEBUG, logger="pgflit"):
        result_apply = instances.apply(ctx, im)
        assert result_apply.change_state == interface.ApplyChangeState.changed
    assert (
        f"instance {i} needs restart due to parameter changes: listen_addresses"
        in caplog.messages
    )
    assert postgresql.status(i) == Status.running
    with connect(i) as conn:
        assert instances.pending_restart(conn)

    # Change runtime state again.
    im = im._copy_validate({"state": "stopped"})
    result_apply = instances.apply(ctx, im)
    assert result_apply.change_state == interface.ApplyChangeState.changed
    assert postgresql.status(i) == Status.not_running

    # Delete.
    im = im._copy_validate({"state": "absent"})
    assert instances.apply(ctx, im).change_state == interface.ApplyChangeState.dropped
    with pytest.raises(exceptions.InstanceNotFound):
        i.exists()
    assert not system.BaseInstance.get("test_apply", pg_version, ctx.settings).exists()


def test_get(
    ctx: Context,
    instance: system.Instance,
    pgbackrest_available: bool,
    powa_available: bool,
) -> None:
    im = instances.get(ctx, instance)
    assert im is not None
    assert im.name == "test"
    config = im.settings
    assert im.port == instance.port
    assert im.data_directory == instance.datadir  # type: ignore[attr-defined]
    assert im.wal_directory == instance.waldir  # type: ignore[attr-defined]
    # Pop host-dependent values.
    del config["effective_cache_size"]
    del config["shared_buffers"]
    spl = "passwordcheck"
    if powa_available:
        spl += ", pg_qualstats, pg_stat_statements, pg_stat_kcache"
    socket_directory = str(ctx.settings.postgresql.socket_directory).format(
        instance=instance
    )
    expected_config = {
        "cluster_name": "test",
        "lc_messages": "C",
        "lc_monetary": "C",
        "lc_numeric": "C",
        "lc_time": "C",
        "log_destination": "stderr",
        "log_directory": str(ctx.settings.postgresql.logpath),
        "log_filename": f"{instance.qualname}-%Y-%m-%d_%H%M%S.log",
        "logging_collector": True,
        "shared_preload_libraries": spl,
        "unix_socket_directories": socket_directory,
    }
    if pgbackrest_available:
        del config["archive_command"]
        expected_config["archive_mode"] = True
        expected_config["wal_level"] = "replica"
    assert config == expected_config
    assert im.data_checksums is False
    assert im.state.name == "started"
    assert not im.pending_restart


def test_ls(settings: Settings, instance: system.Instance) -> None:
    not_instance_dir = Path(
        str(settings.postgresql.datadir).format(version="12", name="notAnInstanceDir")
    )
    not_instance_dir.mkdir(parents=True)
    try:
        ilist = list(instances.ls(settings))

        for i in ilist:
            assert i.status == Status.running.name
            # this also ensure instance name is not notAnInstanceDir
            assert i.name == "test"

        for i in ilist:
            if (i.version, i.name) == (instance.version, instance.name):
                break
        else:
            pytest.fail(f"Instance {instance.version}/{instance.name} not found")

        iv = next(instances.ls(settings, version=instance.version))
        assert iv == i
    finally:
        not_instance_dir.rmdir()


def test_server_settings(instance: system.Instance) -> None:
    with connect(instance) as conn:
        pgsettings = instances.settings(conn)
    port = next(p for p in pgsettings if p.name == "port")
    assert port.setting == str(instance.port)
    assert not port.pending_restart
    assert port.context == "postmaster"


def test_logs(instance: system.Instance) -> None:
    try:
        for line in postgresql.logs(instance, timeout=0):
            if "database system is ready to accept connections" in line:
                break
        else:
            pytest.fail("expected log line not found")
    except TimeoutError:
        pass


def test_get_locale(instance: system.Instance) -> None:
    assert postgresql.is_running(instance)
    with connect(instance) as conn:
        assert instances.get_locale(conn) == "C"


def test_data_checksums(
    ctx: Context,
    pg_version: str,
    instance_factory: Factory[tuple[interface.Instance, system.Instance]],
    caplog: pytest.LogCaptureFixture,
) -> None:
    manifest, instance = instance_factory(ctx.settings, "datachecksums")

    with postgresql.running(ctx, instance):
        assert execute(instance, "SHOW data_checksums") == [{"data_checksums": "off"}]

    # explicitly enabled
    manifest = manifest._copy_validate({"data_checksums": True})
    with caplog.at_level(logging.INFO, logger="pglift.instances"):
        result_apply = instances.apply(ctx, manifest)
        assert result_apply.change_state == interface.ApplyChangeState.changed
    with postgresql.running(ctx, instance):
        assert execute(instance, "SHOW data_checksums") == [{"data_checksums": "on"}]
    assert "enabling data checksums" in caplog.messages
    caplog.clear()

    assert instances._get(ctx, instance).data_checksums

    # not explicitly disabled so still enabled
    result_apply = instances.apply(
        ctx, manifest._copy_validate({"data_checksums": None})
    )
    assert result_apply.change_state is None
    with postgresql.running(ctx, instance):
        assert execute(instance, "SHOW data_checksums") == [{"data_checksums": "on"}]

    # explicitly disabled
    with caplog.at_level(logging.INFO, logger="pglift.instances"):
        result_apply = instances.apply(
            ctx, manifest._copy_validate({"data_checksums": False})
        )
        assert result_apply.change_state == interface.ApplyChangeState.changed
    with postgresql.running(ctx, instance):
        assert execute(instance, "SHOW data_checksums") == [{"data_checksums": "off"}]
    assert "disabling data checksums" in caplog.messages
    caplog.clear()
    assert instances._get(ctx, instance).data_checksums is False

    # re-enabled with instance running
    with postgresql.running(ctx, instance):
        with pytest.raises(
            exceptions.InstanceStateError,
            match="could not alter data_checksums on a running instance",
        ):
            instances.apply(ctx, manifest._copy_validate({"data_checksums": True}))
    assert instances._get(ctx, instance).data_checksums is False



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Protocol


@dataclass
class Certificate:
    path: Path
    private_key: Path

    def __post_init__(self) -> None:
        if not self.path.exists():
            raise ValueError(f"path={self.path} does not exist")
        if not self.private_key.exists():
            raise ValueError(f"private_key={self.private_key} does not exist")


class CertFactory(Protocol):
    def __call__(self, *identities: str, common_name: str | None = None) -> Certificate:
        ...



import grp
import json
import os
import pwd
import tempfile
from pathlib import Path
from typing import Any, Literal, Optional

import pydantic
import yaml
from pydantic import Field, root_validator, validator
from pydantic.env_settings import SettingsSourceCallable
from pydantic.fields import ModelField

from .. import __name__ as pkgname
from .. import exceptions, util
from . import (
    _cli,
    _logrotate,
    _patroni,
    _pgbackrest,
    _postgresql,
    _powa,
    _prometheus,
    _rsyslog,
    _systemd,
    _temboard,
)
from .base import prefix_values


def default_prefix(uid: int) -> Path:
    """Return the default path prefix for 'uid'.

    >>> default_prefix(0)
    PosixPath('/')
    >>> default_prefix(42)  # doctest: +ELLIPSIS
    PosixPath('/.../.local/share/pglift')
    """
    if uid == 0:
        return Path("/")
    return util.xdg_data_home() / pkgname


def default_run_prefix(uid: int) -> Path:
    """Return the default run path prefix for 'uid'."""
    if uid == 0:
        base = Path("/run")
    else:
        try:
            base = util.xdg_runtime_dir(uid)
        except exceptions.FileNotFoundError:
            base = Path(tempfile.gettempdir())

    return base / pkgname


def default_sysuser() -> tuple[str, str]:
    pwentry = pwd.getpwuid(os.getuid())
    grentry = grp.getgrgid(pwentry.pw_gid)
    return pwentry.pw_name, grentry.gr_name


def yaml_settings_source(settings: pydantic.BaseSettings) -> dict[str, Any]:
    """Load settings values 'settings.yaml' file if found in user or system
    config directory directory.
    """
    assert isinstance(settings, SiteSettings)
    path = settings.site_settings()
    if path is None:
        return {}
    settings = yaml.safe_load(path.read_text())
    if not isinstance(settings, dict):
        raise exceptions.SettingsError(
            f"failed to load site settings from {path}, expecting an object"
        )
    return settings


def json_config_settings_source(settings: pydantic.BaseSettings) -> dict[str, Any]:
    """Load settings values from 'SETTINGS' environment variable.

    If this variable has a value starting with @, it is interpreted as a path
    to a JSON file. Otherwise, a JSON serialization is expected.
    """
    env_settings = os.getenv("SETTINGS")
    if not env_settings:
        return {}
    if env_settings.startswith("@"):
        config = Path(env_settings[1:])
        encoding = settings.__config__.env_file_encoding
        # May raise FileNotFoundError, which is okay here.
        env_settings = config.read_text(encoding)
    try:
        return json.loads(env_settings)  # type: ignore[no-any-return]
    except json.decoder.JSONDecodeError as e:
        raise exceptions.SettingsError(str(e)) from e


def is_root() -> bool:
    return os.getuid() == 0


class Settings(pydantic.BaseSettings):
    """Settings for pglift."""

    class Config:
        frozen = True

    postgresql: _postgresql.Settings = Field(default_factory=_postgresql.Settings)
    patroni: Optional[_patroni.Settings] = None
    pgbackrest: Optional[_pgbackrest.Settings] = None
    powa: Optional[_powa.Settings] = None
    prometheus: Optional[_prometheus.Settings] = None
    temboard: Optional[_temboard.Settings] = None
    systemd: Optional[_systemd.Settings] = None
    logrotate: Optional[_logrotate.Settings] = None
    rsyslog: Optional[_rsyslog.Settings] = None
    cli: _cli.Settings = Field(default_factory=_cli.Settings)

    service_manager: Optional[Literal["systemd"]] = None
    scheduler: Optional[Literal["systemd"]] = None

    prefix: Path = Field(
        default=default_prefix(os.getuid()),
        description="Path prefix for configuration and data files.",
    )

    run_prefix: Path = Field(
        default=default_run_prefix(os.getuid()),
        description="Path prefix for runtime socket, lockfiles and PID files.",
    )

    sysuser: tuple[str, str] = Field(
        default_factory=default_sysuser,
        help=(
            "(username, groupname) of system user running PostgreSQL; "
            "mostly applicable when operating PostgreSQL with systemd in non-user mode"
        ),
    )

    @validator("prefix", "run_prefix")
    def __validate_prefix_(cls, value: Path) -> Path:
        """Make sure path settings are absolute."""
        if not value.is_absolute():
            raise ValueError("expecting an absolute path")
        return value

    @root_validator(skip_on_failure=True)
    def __prefix_paths(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Prefix child settings fields with the global 'prefix'."""
        return prefix_values(
            values,
            {"prefix": values["prefix"], "run_prefix": values["run_prefix"]},
        )

    @root_validator(pre=True)
    def __set_service_manager_scheduler(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set 'service_manager' and 'scheduler' to 'systemd' by default if systemd is enabled."""
        if values.get("systemd") is not None:
            values.setdefault("service_manager", "systemd")
            values.setdefault("scheduler", "systemd")
        return values

    @validator("service_manager", "scheduler")
    def __validate_service_manager_scheduler_(
        cls, v: Optional[Literal["systemd"]], values: dict[str, Any]
    ) -> Optional[Literal["systemd"]]:
        """Make sure systemd is enabled globally when 'service_manager' or 'scheduler' are set."""
        if values.get("systemd") is None and v is not None:
            raise ValueError("cannot use systemd, if 'systemd' is not enabled globally")
        return v

    @root_validator(pre=True)
    def __validate_is_not_root(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Make sure current user is not root.

        This is not supported by postgres (cannot call neither initdb nor pg_ctl as root).
        """
        if is_root():
            raise exceptions.UnsupportedError("pglift cannot be used as root")
        return values

    @validator("patroni")
    def __validate_patroni_passfile_(
        cls,
        value: Optional[_patroni.Settings],
        values: dict[str, Any],
        field: ModelField,
    ) -> Optional[_patroni.Settings]:
        try:
            postgresql_settings = values["postgresql"]
        except KeyError:  # Another validation probably failed.
            return value
        assert isinstance(postgresql_settings, _postgresql.Settings)
        if (
            value
            and postgresql_settings.auth.passfile
            and value.postgresql.passfile == postgresql_settings.auth.passfile
        ):
            raise ValueError(
                f"'{field.name}.postgresql.passfile' must be different from 'postgresql.auth.passfile'"
            )
        return value

    @validator("patroni")
    def __validate_patroni_requires_replrole_(
        cls, value: Optional[_patroni.Settings], values: dict[str, Any]
    ) -> Optional[_patroni.Settings]:
        try:
            postgresql_settings = values["postgresql"]
        except KeyError:  # Another validation probably failed.
            return value
        assert isinstance(postgresql_settings, _postgresql.Settings)
        if value and postgresql_settings.replrole is None:
            raise ValueError("'postgresql.replrole' must be provided to use 'patroni'")
        return value


class SiteSettings(Settings):
    """Settings loaded from site-sources.

    Load user or site settings from:
    - 'settings.yaml' if found in user or system configuration directory, and,
    - SETTINGS environment variable.
    """

    @staticmethod
    def site_settings() -> Optional[Path]:
        """Return content of 'settings.yaml' if found in site configuration
        directories.
        """
        for hdlr in (util.xdg_config, util.etc_config):
            if (fpath := hdlr("settings.yaml")) is not None:
                return fpath
        return None

    class Config:
        frozen = True

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> tuple[SettingsSourceCallable, ...]:
            return (
                init_settings,
                env_settings,
                json_config_settings_source,
                yaml_settings_source,
            )



from __future__ import annotations

from pglift import pm, settings


def test_pluginmanager_all() -> None:
    p = pm.PluginManager.all()
    assert {name for name, _ in p.list_name_plugin()} == {
        "pglift.backup",
        "pglift.databases",
        "pglift.logrotate",
        "pglift.passfile",
        "pglift.patroni",
        "pglift.pgbackrest",
        "pglift.pgbackrest.repo_host_ssh",
        "pglift.pgbackrest.repo_host_tls",
        "pglift.pgbackrest.repo_path",
        "pglift.postgresql",
        "pglift.powa",
        "pglift.prometheus",
        "pglift.rsyslog",
        "pglift.systemd.scheduler",
        "pglift.systemd.service_manager",
        "pglift.temboard",
    }


def test_pluginmanager_get(settings: settings.Settings) -> None:
    new_settings = settings.copy(update={"prometheus": None})
    p = pm.PluginManager.get(new_settings)
    assert {name for name, _ in p.list_name_plugin()} == {
        "pglift.databases",
        "pglift.logrotate",
        "pglift.passfile",
        "pglift.patroni",
        "pglift.pgbackrest",
        "pglift.pgbackrest.repo_path",
        "pglift.postgresql",
        "pglift.powa",
        "pglift.rsyslog",
        "pglift.temboard",
    }


def test_pluginmanager_unregister_all(settings: settings.Settings) -> None:
    p = pm.PluginManager.get(settings)
    assert p.list_name_plugin()
    p.unregister_all()
    assert not p.list_name_plugin()


def test_eq(settings: settings.Settings) -> None:
    p1, p2 = pm.PluginManager.get(settings), pm.PluginManager.get(settings)
    assert p1 is not p2
    assert p1 == p2

    p2.unregister_all()
    assert p1 != p2

    assert p1 != object()
    assert 42 != p2



from pydantic import Field

from .base import BaseModel, ConfigPath


class Settings(BaseModel):
    """Settings for logrotate."""

    configdir: ConfigPath = Field(
        default=ConfigPath("logrotate.d"), description="Logrotate config directory"
    )



import enum
from pathlib import Path
from typing import Literal

from pydantic import AnyHttpUrl, Field, FilePath

from .. import types
from .base import BaseModel, ConfigPath, DataPath, LogPath, RunPath, ServerCert


class Plugin(types.AutoStrEnum):
    activity = enum.auto()
    administration = enum.auto()
    dashboard = enum.auto()
    maintenance = enum.auto()
    monitoring = enum.auto()
    pgconf = enum.auto()
    statements = enum.auto()


class LogMethod(types.AutoStrEnum):
    stderr = enum.auto()
    syslog = enum.auto()
    file = enum.auto()


class Settings(BaseModel):
    """Settings for temBoard agent"""

    ui_url: AnyHttpUrl = Field(description="URL of the temBoard UI.")

    signing_key: FilePath = Field(
        description="Path to the public key for UI connection."
    )

    certificate: ServerCert = Field(
        description="TLS certificate files for the temboard-agent HTTP server."
    )

    execpath: FilePath = Field(
        default=Path("/usr/bin/temboard-agent"),
        description="Path to the temboard-agent executable.",
    )

    role: str = Field(
        default="temboardagent",
        description="Name of the PostgreSQL role for temBoard agent.",
    )

    configpath: ConfigPath = Field(
        default=ConfigPath("temboard-agent/temboard-agent-{name}.conf"),
        description="Path to the config file.",
    )

    pid_file: RunPath = Field(
        default=RunPath("temboard-agent/temboard-agent-{name}.pid"),
        description="Path to which temboard-agent process PID will be written.",
    )

    plugins: tuple[Plugin, ...] = Field(
        default=(
            Plugin.monitoring,
            Plugin.dashboard,
            Plugin.activity,
        ),
        description="Plugins to load.",
    )

    home: DataPath = Field(
        default=DataPath("temboard-agent/{name}"),
        description="Path to agent home directory containing files used to store temporary data",
    )

    logpath: LogPath = Field(
        default=LogPath("temboard"),
        description="Path where log files are stored.",
    )

    logmethod: LogMethod = Field(
        default=LogMethod.stderr, description="Method used to send the logs."
    )

    loglevel: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = Field(
        default="INFO", description="Log level."
    )



from __future__ import annotations

import logging
from collections.abc import Iterator
from typing import Any

from pgtoolkit.conf import Configuration
from pydantic import Field

from .. import hookimpl
from .. import service as service_mod
from .. import systemd, util
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings
from ..types import Status
from . import impl, models
from .impl import available as available
from .impl import get_settings

logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    return available(settings) is not None


@hookimpl
def site_configure_install(settings: Settings) -> None:
    s = get_settings(settings)
    logger.info("creating temboard log directory")
    s.logpath.mkdir(mode=0o740, parents=True, exist_ok=True)


@hookimpl
def site_configure_uninstall(settings: Settings) -> None:
    s = get_settings(settings)
    logger.info("deleting temboard log directory")
    util.rmtree(s.logpath)


@hookimpl
def site_configure_installed(settings: Settings) -> bool:
    s = get_settings(settings)
    if not s.logpath.exists():
        logger.error("temBoard log directory '%s' missing", s.logpath)
        return False
    return True


@hookimpl
def system_lookup(instance: system.PostgreSQLInstance) -> models.Service | None:
    settings = get_settings(instance._settings)
    return impl.system_lookup(instance.qualname, settings)


@hookimpl
def interface_model() -> tuple[str, Any, Any]:
    return (
        models.ServiceManifest.__service__,
        models.ServiceManifest,
        Field(
            default=models.ServiceManifest(),
            description="Configuration for the temBoard service, if enabled in site settings.",
        ),
    )


@hookimpl
def get(instance: system.Instance) -> models.ServiceManifest | None:
    try:
        s = instance.service(models.Service)
    except ValueError:
        return None
    else:
        return models.ServiceManifest(port=s.port)


SYSTEMD_SERVICE_NAME = "pglift-temboard_agent@.service"


@hookimpl
def systemd_unit_templates(
    settings: Settings, content: bool
) -> Iterator[str] | Iterator[tuple[str, str]]:
    if not content:
        yield SYSTEMD_SERVICE_NAME
        return
    s = get_settings(settings)
    configpath = str(s.configpath).replace("{name}", "%i")
    yield SYSTEMD_SERVICE_NAME, systemd.template(SYSTEMD_SERVICE_NAME).format(
        executeas=systemd.executeas(settings),
        configpath=configpath,
        execpath=str(s.execpath),
    )


@hookimpl
def instance_configured(
    ctx: Context, manifest: interface.Instance, config: Configuration
) -> None:
    """Install temboard agent for an instance when it gets configured."""
    settings = get_settings(ctx.settings)
    impl.setup(ctx, manifest, settings, config)


@hookimpl
def instance_started(ctx: Context, instance: system.Instance) -> None:
    """Start temboard agent service."""
    try:
        service = instance.service(models.Service)
    except ValueError:
        return
    impl.start(ctx, service)


@hookimpl
def instance_stopped(ctx: Context, instance: system.Instance) -> None:
    """Stop temboard agent service."""
    try:
        service = instance.service(models.Service)
    except ValueError:
        return
    impl.stop(ctx, service)


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    """Uninstall temboard from an instance being dropped."""
    settings = get_settings(instance._settings)
    manifest = interface.Instance(name=instance.name, version=instance.version)
    impl.revert_setup(ctx, manifest, settings, instance.config())


@hookimpl
def rolename(settings: Settings) -> str:
    assert settings.temboard
    return settings.temboard.role


@hookimpl
def role(settings: Settings, manifest: interface.Instance) -> interface.Role:
    name = rolename(settings)
    service_manifest = manifest.service_manifest(models.ServiceManifest)
    return interface.Role(
        name=name, password=service_manifest.password, login=True, superuser=True
    )


@hookimpl
def instance_status(
    ctx: Context, instance: system.Instance
) -> tuple[Status, str] | None:
    try:
        service = instance.service(models.Service)
    except ValueError:
        return None
    return (service_mod.status(ctx, service), "temBoard")



from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import patch

import psycopg.rows
import pytest
from psycopg import sql

from pglift import db
from pglift.ctx import Context
from pglift.models.system import Instance
from pglift.settings import Settings


def test_queries(datadir: Path, write_changes: bool) -> None:
    actual = dict(db.queries())
    fpath = datadir / "queries.json"
    if write_changes:
        fpath.write_text(json.dumps(actual, indent=2, sort_keys=True) + "\n")
    expected = json.loads(fpath.read_text())
    assert actual == expected


def test_query() -> None:
    query = db.query(
        "role_alter",
        username=sql.Identifier("bob"),
        options=sql.Literal("PASSWORD 'ha'"),
    )
    assert list(query) == [
        sql.SQL("ALTER ROLE "),
        sql.Identifier("bob"),
        sql.SQL(" "),
        sql.Literal("PASSWORD 'ha'"),
        sql.SQL(";"),
    ]


@pytest.mark.parametrize(
    "connargs, expected",
    [
        (
            {"user": "bob"},
            "dbname=mydb sslmode=off user=bob port=999 host=/socks passfile={passfile}",
        ),
        (
            {"user": "alice", "password": "s3kret"},
            "dbname=mydb sslmode=off user=alice password=s3kret port=999 host=/socks passfile={passfile}",
        ),
    ],
)
def test_dsn(
    settings: Settings, instance: Instance, connargs: dict[str, str], expected: str
) -> None:
    passfile = settings.postgresql.auth.passfile
    conninfo = db.dsn(instance, dbname="mydb", sslmode="off", **connargs)
    assert conninfo == expected.format(passfile=passfile)


def test_dsn_badarg(instance: Instance) -> None:
    with pytest.raises(TypeError, match="unexpected 'port' argument"):
        db.dsn(instance, port=123)


def test_connect_str() -> None:
    with patch("psycopg.connect", autospec=True) as connect:
        with db.connect("port=1234"):
            pass
    connect.assert_called_once_with(
        "port=1234", autocommit=True, row_factory=psycopg.rows.dict_row
    )


def test_connect_instance(ctx: Context, instance: Instance, settings: Settings) -> None:
    with patch("psycopg.connect", autospec=True) as connect:
        cnx = db.connect(instance, ctx=ctx, user="dba")
        assert not connect.called
        with cnx:
            pass
    passfile = settings.postgresql.auth.passfile
    assert passfile is not None and passfile.exists()
    connect.assert_called_once_with(
        f"user=dba port=999 host=/socks passfile={passfile}",
        autocommit=True,
        row_factory=psycopg.rows.dict_row,
    )


def test_primary_connect(standby_instance: Instance, settings: Settings) -> None:
    standby = standby_instance.standby
    assert standby
    with patch("psycopg.connect", autospec=True) as connect:
        with db.primary_connect(standby):
            pass
    connect.assert_called_once_with(
        "user=pg host=/tmp port=4242 dbname=template1",
        autocommit=True,
        row_factory=psycopg.rows.dict_row,
    )


def test_connect_notimplementederror() -> None:
    with pytest.raises(NotImplementedError):
        db.connect({"host": "localhost"})



import socket
from typing import ClassVar, Optional

from pydantic import Extra, Field, FilePath, SecretStr, validator

from ... import types
from .common import RESTAPI


class ClusterMember(types.BaseModel, extra=Extra.allow, frozen=True):
    """An item of the list of members returned by Patroni API /cluster endpoint."""

    host: str
    name: str
    port: int
    role: str
    state: str


class ClientSSLOptions(types.BaseModel):
    cert: FilePath = Field(description="Client certificate.")
    key: FilePath = Field(description="Private key.")
    password: Optional[SecretStr] = Field(
        default=None, description="Password for the private key."
    )


class ClientAuth(types.BaseModel):
    ssl: Optional[ClientSSLOptions] = Field(
        default=None,
        description="Client certificate options.",
    )


class PostgreSQL(types.BaseModel):
    connect_host: Optional[str] = Field(
        default=None,
        description="Host or IP address through which PostgreSQL is externally accessible.",
    )
    replication: Optional[ClientAuth] = Field(
        default=None,
        description="Authentication options for client (libpq) connections to remote PostgreSQL by the replication user.",
    )
    rewind: Optional[ClientAuth] = Field(
        default=None,
        description="Authentication options for client (libpq) connections to remote PostgreSQL by the rewind user.",
    )


class Etcd(types.BaseModel):
    username: str = Field(
        description="Username for basic authentication to etcd.",
    )
    password: SecretStr = Field(
        description="Password for basic authentication to etcd."
    )


class ServiceManifest(types.ServiceManifest, service_name="patroni"):
    _cli_config: ClassVar[dict[str, types.CLIConfig]] = {
        "cluster_members": {"hide": True},
        "postgresql_connect_host": {"hide": True},
    }
    _ansible_config: ClassVar[dict[str, types.AnsibleConfig]] = {
        "cluster_members": {"hide": True},
    }

    # XXX Or simply use instance.qualname?
    cluster: str = Field(
        description="Name (scope) of the Patroni cluster.",
        readOnly=True,
    )
    node: str = Field(
        default_factory=socket.getfqdn,
        description="Name of the node (usually the host name).",
        readOnly=True,
    )
    restapi: RESTAPI = Field(
        default_factory=RESTAPI, description="REST API configuration"
    )

    postgresql: Optional[PostgreSQL] = Field(
        default=None,
        description="Configuration for PostgreSQL setup and remote connection.",
    )
    etcd: Optional[Etcd] = Field(
        default=None, description="Instance-specific options for etcd DCS backend."
    )
    cluster_members: list[ClusterMember] = Field(
        default=[],
        description="Members of the Patroni this instance is member of.",
        readOnly=True,
    )

    __validate_none_values_ = validator("node", "restapi", pre=True, allow_reuse=True)(
        types.default_if_none
    )



from __future__ import annotations

import logging
from typing import Any

import psycopg.pq
import psycopg.rows
from psycopg import sql

from . import databases, db, exceptions
from .ctx import Context
from .models import interface, system
from .types import Role

logger = logging.getLogger(__name__)


def apply(
    ctx: Context, instance: system.PostgreSQLInstance, role: interface.Role
) -> interface.ApplyResult:
    """Apply state described by specified interface model as a PostgreSQL role.

    In case it's not possible to inspect changed role, possibly due to the super-user
    password being modified, change_state attribute within the returned object
    is set to interface.ApplyResult.changed with a warning logged.

    The instance should be running and not a standby.

    :raises ~pglift.exceptions.DependencyError: if the role cannot be dropped due some database dependency.
    """
    if instance.standby:
        raise exceptions.InstanceReadOnlyError(instance)

    with db.connect(instance, ctx=ctx) as cnx:
        return _apply(cnx, role, instance, ctx)


def _apply(
    cnx: db.Connection,
    role: interface.Role,
    instance: system.PostgreSQLInstance,
    ctx: Context,
) -> interface.ApplyResult:
    name = role.name
    if role.state == interface.PresenceState.absent:
        dropped = False
        if _exists(cnx, name):
            _drop(cnx, role, instance=instance, ctx=ctx)
            dropped = True
        return interface.ApplyResult(
            change_state=interface.ApplyChangeState.dropped if dropped else None
        )

    if not _exists(cnx, name):
        _create(cnx, role)
        ctx.hook.role_change(role=role, instance=instance)
        return interface.ApplyResult(change_state=interface.ApplyChangeState.created)
    else:
        actual = _get(cnx, name, instance=instance, ctx=ctx, password=False)
        _alter(cnx, role, instance=instance, ctx=ctx)
        if any(ctx.hook.role_change(role=role, instance=instance)):
            return interface.ApplyResult(
                change_state=interface.ApplyChangeState.changed
            )
        changed = _get(cnx, name, instance=instance, ctx=ctx, password=False) != actual
        return interface.ApplyResult(
            change_state=interface.ApplyChangeState.changed if changed else None
        )


def get(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    name: str,
    *,
    password: bool = True,
) -> interface.Role:
    """Return the role object with specified name.

    :raises ~pglift.exceptions.RoleNotFound: if no role with specified 'name' exists.
    """
    with db.connect(instance, ctx=ctx) as cnx:
        return _get(cnx, name, instance=instance, ctx=ctx, password=password)


def _get(
    cnx: db.Connection,
    name: str,
    *,
    instance: system.PostgreSQLInstance,
    ctx: Context,
    password: bool = True,
) -> interface.Role:
    values = cnx.execute(db.query("role_inspect"), {"username": name}).fetchone()
    if values is None:
        raise exceptions.RoleNotFound(name)
    if not password:
        values["password"] = None
    for extra in ctx.hook.role_inspect(instance=instance, name=name):
        conflicts = set(values) & set(extra)
        assert (
            not conflicts
        ), f"conflicting keys returned by role_inspect() hook: {', '.join(conflicts)}"
        values.update(extra)
    return interface.Role(**values)


def ls(ctx: Context, instance: system.PostgreSQLInstance) -> list[interface.Role]:
    """Return the list of roles for an instance."""
    with db.connect(instance, ctx=ctx) as cnx, cnx.cursor(
        row_factory=psycopg.rows.class_row(interface.Role)
    ) as cur:
        return cur.execute(db.query("role_list")).fetchall()


def drop(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    role: interface.RoleDropped,
) -> None:
    """Drop a role from instance.

    :raises ~pglift.exceptions.RoleNotFound: if no role with specified 'role.name' exists.
    :raises ~pglift.exceptions.RoleNotFound: if no role with specified 'role.reassign_owned' exists.
    :raises ~pglift.exceptions.DependencyError: if the role cannot be dropped due some database dependency.
    """
    if instance.standby:
        raise exceptions.InstanceReadOnlyError(instance)
    with db.connect(instance, ctx=ctx) as cnx:
        if not _exists(cnx, role.name):
            raise exceptions.RoleNotFound(role.name)
        _drop(cnx, role, instance=instance, ctx=ctx)


def _drop(
    cnx: db.Connection,
    role: interface.RoleDropped,
    *,
    instance: system.PostgreSQLInstance,
    ctx: Context,
) -> None:
    if role.reassign_owned and not _exists(cnx, role.reassign_owned):
        raise exceptions.RoleNotFound(role.reassign_owned)
    logger.info("dropping role '%s'", role.name)
    dbs_to_drop: list[str] = []
    if role.drop_owned or role.reassign_owned:
        for database in databases._list(cnx):
            if role.drop_owned and database.owner == role.name:
                dbs_to_drop.append(database.name)
            else:
                if role.drop_owned:
                    query = db.query(
                        "role_drop_owned", username=sql.Identifier(role.name)
                    )
                elif role.reassign_owned:
                    query = db.query(
                        "role_drop_reassign",
                        oldowner=sql.Identifier(role.name),
                        newowner=sql.Identifier(role.reassign_owned),
                    )
                with db.connect(instance, ctx=ctx, dbname=database.name) as db_cnx:
                    db_cnx.execute(query)

    for dbname in dbs_to_drop:
        cnx.execute(
            db.query(
                "database_drop",
                database=sql.Identifier(dbname),
                options=sql.SQL(""),
            )
        )
    try:
        cnx.execute(db.query("role_drop", username=sql.Identifier(role.name)))
    except psycopg.errors.DependentObjectsStillExist as e:
        assert (
            not role.drop_owned and not role.reassign_owned
        ), f"unexpected {e!r} while dropping {role}"
        raise exceptions.DependencyError(
            f"{e.diag.message_primary} (detail: {e.diag.message_detail})"
        ) from e

    ctx.hook.role_change(role=role, instance=instance)


def exists(ctx: Context, instance: system.PostgreSQLInstance, name: str) -> bool:
    """Return True if named role exists in 'instance'.

    The instance should be running.
    """
    with db.connect(instance, ctx=ctx) as cnx:
        return _exists(cnx, name)


def _exists(cnx: db.Connection, name: str) -> bool:
    cur = cnx.execute(db.query("role_exists"), {"username": name})
    return cur.rowcount == 1


def encrypt_password(cnx: psycopg.Connection[Any], role: Role) -> str:
    if role.encrypted_password is not None:
        return role.encrypted_password.get_secret_value()
    assert role.password is not None, "role has no password to encrypt"
    encoding = cnx.info.encoding
    return cnx.pgconn.encrypt_password(
        role.password.get_secret_value().encode(encoding), role.name.encode(encoding)
    ).decode(encoding)


def options(
    cnx: psycopg.Connection[Any],
    role: interface.Role,
    *,
    in_roles: bool = True,
) -> sql.Composable:
    """Return the "options" part of CREATE ROLE or ALTER ROLE SQL commands
    based on 'role' model.
    """
    opts: list[sql.Composable] = [
        sql.SQL("INHERIT" if role.inherit else "NOINHERIT"),
        sql.SQL("LOGIN" if role.login else "NOLOGIN"),
        sql.SQL("SUPERUSER" if role.superuser else "NOSUPERUSER"),
        sql.SQL("REPLICATION" if role.replication else "NOREPLICATION"),
        sql.SQL("CREATEDB" if role.createdb else "NOCREATEDB"),
        sql.SQL("CREATEROLE" if role.createrole else "NOCREATEROLE"),
    ]
    if role.password or role.encrypted_password:
        opts.append(sql.SQL("PASSWORD {}").format(encrypt_password(cnx, role)))
    if role.validity is not None:
        opts.append(sql.SQL("VALID UNTIL {}").format(role.validity.isoformat()))
    opts.append(
        sql.SQL(
            "CONNECTION LIMIT {}".format(
                role.connection_limit if role.connection_limit is not None else -1
            )
        )
    )
    if in_roles and role.in_roles:
        opts.append(
            sql.SQL(" ").join(
                [
                    sql.SQL("IN ROLE"),
                    sql.SQL(", ").join(
                        sql.Identifier(in_role) for in_role in role.in_roles
                    ),
                ]
            )
        )
    return sql.SQL(" ").join(opts)


def _create(cnx: db.Connection, role: interface.Role) -> None:
    logger.info("creating role '%s'", role.name)
    opts = options(cnx, role)
    cnx.execute(
        db.query("role_create", username=sql.Identifier(role.name), options=opts)
    )


def _alter(
    cnx: db.Connection,
    role: interface.Role,
    *,
    instance: system.PostgreSQLInstance,
    ctx: Context,
) -> None:
    logger.info("altering role '%s'", role.name)
    actual_role = _get(cnx, role.name, instance=instance, ctx=ctx)
    in_roles = {
        "grant": set(role.in_roles) - set(actual_role.in_roles),
        "revoke": set(actual_role.in_roles) - set(role.in_roles),
    }
    with cnx.transaction():
        opts = options(cnx, role, in_roles=False)
        cnx.execute(
            db.query(
                "role_alter",
                username=sql.Identifier(role.name),
                options=opts,
            ),
        )
        for action, values in in_roles.items():
            if values:
                cnx.execute(
                    db.query(
                        f"role_{action}",
                        rolname=sql.SQL(", ").join(sql.Identifier(r) for r in values),
                        rolspec=sql.Identifier(role.name),
                    )
                )



from __future__ import annotations

from pglift import rsyslog
from pglift.pm import PluginManager
from pglift.settings import Settings


def test_site_configure(settings: Settings, pm: PluginManager) -> None:
    rsyslog_settings = settings.rsyslog
    assert rsyslog_settings is not None
    rsyslog_config_file = rsyslog_settings.configdir / "rsyslog.conf"
    assert not rsyslog_config_file.exists()

    assert not rsyslog.site_configure_installed(settings)

    rsyslog.site_configure_install(settings, pm)
    assert rsyslog_settings.configdir.exists()
    assert rsyslog_config_file.exists()
    assert rsyslog.site_configure_installed(settings)

    username, group = settings.sysuser
    assert rsyslog_config_file.read_text().strip() == "\n".join(
        [
            "$umask 0027",
            "$FileCreateMode 0640",
            f"$FileOwner {username}",
            f"$FileGroup {group}",
            f'template (name="pglift_postgresql_template" type="string" string="{settings.postgresql.logpath}/%PROGRAMNAME%.log")',
            'if (re_match($programname, "postgresql-.*")) then -?pglift_postgresql_template',
            "&stop",
        ]
    )

    rsyslog.site_configure_uninstall(settings=settings)
    assert not rsyslog_config_file.exists()



from __future__ import annotations

import abc
import enum
import json
import logging
import os
import pathlib
import tempfile
import time
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from functools import cache, cached_property, singledispatch, wraps
from types import ModuleType
from typing import Any, Callable, Generic, TypeVar

import click
import filelock
import humanize
import psycopg
import pydantic
import pydantic.json
import rich
from click.shell_completion import CompletionItem
from rich.table import Table
from typing_extensions import ParamSpec

from .. import __name__ as pkgname
from .. import _install, exceptions, instances, plugin_manager, task
from ..ctx import Context
from ..models import helpers, system
from ..settings import Settings, SiteSettings
from ..settings._postgresql import PostgreSQLVersion
from ..task import Displayer
from ..types import AutoStrEnum, ByteSize, CompositeManifest

logger = logging.getLogger(pkgname)


@singledispatch
def prettify(value: Any) -> str:
    """Prettify a value

    The prettification will depend on value type.

    >>> prettify(ByteSize(1024))
    '1.0 kB'
    >>> prettify([None, 1, "foo"])
    'None, 1, foo'
    >>> prettify({"z", "b", "a"})
    'a, b, z'
    >>> prettify(None)
    ''
    >>> prettify({"foo": "bob"})
    'foo: bob'
    """
    return str(value)


@prettify.register(ByteSize)
def _(value: ByteSize) -> str:
    """Prettify a ByteSize value"""
    return humanize.naturalsize(value)


@prettify.register(list)
def _(value: list[Any]) -> str:
    """Prettify a List value"""
    return ", ".join(str(x) for x in value)


@prettify.register(set)
def _(value: set[Any]) -> str:
    """Prettify a Set value"""
    return prettify(sorted(value))


@prettify.register(type(None))
def _(value: None) -> str:
    """Prettify a None value"""
    return ""


@prettify.register(dict)
def _(value: dict[str, Any]) -> str:
    """Prettify a Dict value.

    >>> print(prettify({"foo": "bob", "bar": {"blah": ["some", 123]}}))
    foo: bob
    bar:
      blah: some, 123
    """

    def prettify_dict(
        d: dict[str, Any], level: int = 0, indent: str = "  "
    ) -> Iterator[str]:
        for key, value in d.items():
            row = f"{indent * level}{key}:"
            if isinstance(value, dict):
                yield row
                yield from prettify_dict(value, level + 1)
            else:
                yield row + " " + prettify(value)

    return "\n".join(prettify_dict(value))


def print_table_for(
    items: Iterable[Mapping[str, Any]],
    title: str | None = None,
    **kwargs: Any,
) -> None:
    """Render a list of items as a table.

    >>> items = [dict(name="bob", children=["marley", "dylan"],
    ...               address=dict(street="main street", zip=31234, city="luz")),
    ...          dict(name="janis", children=None,
    ...               address=dict(street="robinson lane", zip=38650, city="mars"))]
    >>> print_table_for(items, title="address book")  # doctest: +NORMALIZE_WHITESPACE
                      address book
    ┏━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ name  ┃ children      ┃ address               ┃
    ┡━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━┩
    │ bob   │ marley, dylan │ street: main street   │
    │       │               │ zip: 31234            │
    │       │               │ city: luz             │
    │ janis │               │ street: robinson lane │
    │       │               │ zip: 38650            │
    │       │               │ city: mars            │
    └───────┴───────────────┴───────────────────────┘
    """
    table = None
    headers: list[str] = []
    rows = []
    for item in items:
        row = []
        hdr = []
        for k, v in list(item.items()):
            hdr.append(k)
            row.append(prettify(v))
        if not headers:
            headers = hdr[:]
        rows.append(row)
    if not rows:
        return
    table = Table(*headers, title=title, **kwargs)
    for row in rows:
        table.add_row(*row)
    rich.print(table)


def print_json_for(data: Any) -> None:
    """Render `data` as JSON.

    >>> class Foo(pydantic.BaseModel):
    ...     bar_: str = pydantic.Field(alias="bar")
    ...     baz: int
    >>> items = [Foo(bar="x", baz=1), Foo(bar="y", baz=3)]
    >>> print_json_for(items)
    [
      {
        "bar_": "x",
        "baz": 1
      },
      {
        "bar_": "y",
        "baz": 3
      }
    ]
    >>> print_json_for(items[0].dict(by_alias=True))
    {
      "bar": "x",
      "baz": 1
    }
    """
    rich.print_json(json.dumps(data, default=pydantic.json.pydantic_encoder))


P = ParamSpec("P")


def print_schema(
    context: click.Context,
    param: click.Parameter,
    value: bool,
    *,
    model: type[pydantic.BaseModel],
) -> None:
    """Callback for --schema flag."""
    if value:
        rich.print_json(model.schema_json(indent=2))
        context.exit()


def print_argspec(
    context: click.Context,
    param: click.Parameter,
    value: bool,
    *,
    model: type[pydantic.BaseModel],
) -> None:
    """Callback for --ansible-argspec flag."""
    if value:
        click.echo(
            json.dumps(helpers.argspec_from_model(model), sort_keys=False, indent=2)
        )
        context.exit()


def pass_ctx(f: Callable[P, None]) -> Callable[P, None]:
    """Command decorator passing 'Context' bound to click.Context's object."""

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        context = click.get_current_context()
        ctx: Context = context.obj.ctx
        assert isinstance(ctx, Context), ctx
        context.invoke(f, ctx, *args, **kwargs)

    return wrapper


def pass_instance(f: Callable[P, None]) -> Callable[P, None]:
    """Command decorator passing 'instance' bound to click.Context's object."""

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        context = click.get_current_context()
        instance = context.obj.instance
        assert isinstance(instance, system.Instance), instance
        context.invoke(f, instance, *args, **kwargs)

    return wrapper


def pass_component_settings(
    mod: ModuleType, name: str, f: Callable[P, None]
) -> Callable[P, None]:
    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        context = click.get_current_context()
        ctx: Context = context.obj.ctx
        assert isinstance(ctx, Context), ctx
        settings = mod.available(ctx.settings)
        assert settings
        context.invoke(f, settings, *args, **kwargs)

    return wrapper


def get_instance(name: str, version: str | None, settings: Settings) -> system.Instance:
    """Return an Instance from name/version, possibly guessing version if unspecified."""
    if version is None:
        found = None
        for version in PostgreSQLVersion:
            try:
                instance = system.Instance.system_lookup((name, version, settings))
            except exceptions.InstanceNotFound:
                logger.debug("instance '%s' not found in version %s", name, version)
            else:
                if found:
                    raise click.BadParameter(
                        f"instance {name!r} exists in several PostgreSQL versions;"
                        " please select version explicitly"
                    )
                found = instance

        if found:
            return found

        raise click.BadParameter(f"instance {name!r} not found")

    try:
        return system.Instance.system_lookup((name, version, settings))
    except Exception as e:
        raise click.BadParameter(str(e)) from None


def nameversion_from_id(instance_id: str) -> tuple[str, str | None]:
    version = None
    try:
        version, name = instance_id.split("/", 1)
    except ValueError:
        name = instance_id
    return name, version


def instance_lookup(
    context: click.Context, param: click.Parameter, value: None | str | tuple[str]
) -> system.Instance | tuple[system.Instance, ...]:
    """Return one or more system.Instance, possibly guessed if there is only
    one on system, depending on 'param' variadic flag (nargs).
    """

    ctx: Context = context.obj.ctx
    settings = ctx.settings

    def guess() -> tuple[str, str | None]:
        """Return (name, version) of the instance found on system, if there's
        only one, or fail.
        """
        try:
            (i,) = instances.system_list(settings)
        except ValueError:
            raise click.UsageError(
                f"argument {param.get_error_hint(context)} is required."
            ) from None
        return i.name, i.version

    if context.params.get("all_instances"):
        return tuple(
            get_instance(i.name, i.version, i._settings)
            for i in instances.system_list(settings)
        )

    if param.nargs == 1:
        if value is None:
            name, version = guess()
        else:
            assert isinstance(value, str)
            name, version = nameversion_from_id(value)
        return get_instance(name, version, settings)

    elif param.nargs == -1:
        assert isinstance(value, tuple)
        if value:
            return tuple(
                get_instance(*nameversion_from_id(item), settings) for item in value
            )
        else:
            name, version = guess()
            return (get_instance(name, version, settings),)

    else:
        raise AssertionError(f"unexpected nargs={param.nargs}")


def instance_bind_context(
    context: click.Context, param: click.Parameter, value: str | None
) -> None:
    """Bind instance specified as -i/--instance to context's object, possibly
    guessing from available instance if there is only one.
    """
    obj: Obj = context.obj
    settings = obj.ctx.settings
    version: str | None
    if value is None:
        values = list(instances.ls(settings))
        if not values:
            obj._instance = "no instance found; create one first."
            return
        elif len(values) > 1:
            option = param.get_error_hint(context)
            obj._instance = f"several instances found; option {option} is required."
            return
        (i,) = values
        name, version = i.name, i.version
    else:
        name, version = nameversion_from_id(value)
    instance = get_instance(name, version, settings)
    obj._instance = instance


def _list_instances(
    context: click.Context, param: click.Parameter, incomplete: str
) -> list[CompletionItem]:
    """Shell completion function for instance identifier <name> or <version>/<name>."""
    out = []
    iname, iversion = nameversion_from_id(incomplete)
    settings = SiteSettings()
    for i in instances.system_list(settings):
        if iversion is not None and i.version.startswith(iversion):
            if i.name.startswith(iname):
                out.append(
                    CompletionItem(f"{i.version}/{i.name}", help=f"port={i.port}")
                )
            else:
                out.append(CompletionItem(i.version))
        else:
            out.append(
                CompletionItem(i.name, help=f"{i.version}/{i.name} port={i.port}")
            )
    return out


instance_identifier_option = click.option(
    "-i",
    "--instance",
    "instance",
    metavar="<version>/<name>",
    callback=instance_bind_context,
    shell_complete=_list_instances,
    help=(
        "Instance identifier; the <version>/ prefix may be omitted if "
        "there's only one instance matching <name>. "
        "Required if there is more than one instance on system."
    ),
)


class OutputFormat(AutoStrEnum):
    """Output format"""

    json = enum.auto()


output_format_option = click.option(
    "-o",
    "--output-format",
    type=click.Choice(helpers.choices_from_enum(OutputFormat), case_sensitive=False),
    help="Specify the output format.",
)

dry_run_option = click.option(
    "--dry-run", is_flag=True, help="Only validate input data."
)


def validate_foreground(
    context: click.Context, param: click.Parameter, value: bool
) -> bool:
    ctx: Context = context.obj.ctx
    if ctx.settings.service_manager == "systemd" and value:
        raise click.BadParameter("cannot be used with systemd")
    return value


foreground_option = click.option(
    "--foreground",
    is_flag=True,
    help="Start the program in foreground.",
    callback=validate_foreground,
)


@contextmanager
def command_logging(logdir: pathlib.Path) -> Iterator[None]:
    logdir_created = False
    logfilename = f"{time.time()}.log"
    logfile = logdir / logfilename
    try:
        if not logdir.exists():
            logdir.mkdir(parents=True)
            logdir_created = True
        handler = logging.FileHandler(logfile)
    except OSError:
        # Might be, e.g. PermissionError, if log file path is not writable.
        logfile = pathlib.Path(
            tempfile.NamedTemporaryFile(prefix="pglift", suffix=logfilename).name
        )
        handler = logging.FileHandler(logfile)
    formatter = logging.Formatter(
        fmt="%(levelname)-8s - %(asctime)s - %(name)s:%(filename)s:%(lineno)d - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    if logdir_created:
        logger.debug("created log directory %s", logdir)
    logger.debug("logging command at %s", logfile)
    keep_logfile = False
    try:
        yield None
    except (click.Abort, click.ClickException, click.exceptions.Exit):
        raise
    except Exception:
        keep_logfile = True
        logger.exception("an unexpected error occurred")
        raise click.ClickException(
            "an unexpected error occurred, this is probably a bug; "
            f"details can be found at {logfile}"
        ) from None
    finally:
        if not keep_logfile:
            os.unlink(logfile)
            if logdir_created and next(logdir.iterdir(), None) is None:
                logger.debug("removing log directory %s", logdir)
                logdir.rmdir()


class InvalidSettingsError(click.ClickException):
    """Failure to load site settings."""

    def __init__(self, error: Exception) -> None:
        message = f"invalid site settings\n{error}"
        super().__init__(message)


class LogDisplayer:
    def handle(self, msg: str) -> None:
        logger.info(msg)


class CLIContext(Context):
    """Default CLI context, non-interactive."""

    def confirm(self, message: str, default: bool) -> bool:
        return default


class InteractiveCLIContext(CLIContext):
    """An interactive CLI context that prompts for confirmation."""

    def confirm(self, message: str, default: bool) -> bool:
        return rich.prompt.Confirm().ask(
            f"[yellow]>[/yellow] {message}", default=default
        )

    @cache
    def prompt(self, message: str, hide_input: bool = False) -> str:
        return rich.prompt.Prompt().ask(
            f"[yellow]>[/yellow] {message}", password=hide_input
        )


class Obj:
    """Object bound to click.Context"""

    # Set in commands taking a -i/--instance option through
    # instance_identifier_option decorator's callback.
    _instance: str | system.Instance

    def __init__(
        self,
        *,
        context: CLIContext | None = None,
        displayer: Displayer | None = None,
        interactive: bool = True,
        debug: bool = False,
    ) -> None:
        if context is None:
            cls = InteractiveCLIContext if interactive else CLIContext
            try:
                settings = SiteSettings()
            except (exceptions.SettingsError, pydantic.ValidationError) as e:
                raise InvalidSettingsError(e) from e
            except exceptions.UnsupportedError as e:
                raise click.ClickException(f"unsupported operation: {e}") from None
            context = cls(settings=settings)
        self.ctx = context
        self.displayer = displayer
        self.debug = debug

    @staticmethod
    def default_settings() -> Settings:
        return Settings()

    @cached_property
    def instance(self) -> system.Instance:
        if isinstance(self._instance, system.Instance):
            return self._instance
        else:
            assert isinstance(self._instance, str)
            raise click.UsageError(self._instance)


class Command(click.Command):
    def invoke(self, context: click.Context) -> Any:
        obj: Obj = context.obj
        ctx = obj.ctx
        displayer = context.obj.displayer
        logger = logging.getLogger(pkgname)
        with command_logging(ctx.settings.cli.logpath):
            try:
                with task.displayer_installed(displayer):
                    return super().invoke(context)
            except filelock.Timeout:
                raise click.ClickException("another operation is in progress") from None
            except exceptions.Cancelled as e:
                logger.warning(str(e))
                raise click.Abort from None
            except exceptions.Error as e:
                logger.debug("an internal error occurred", exc_info=obj.debug)
                msg = str(e)
                if isinstance(e, exceptions.CommandError):
                    if e.stderr:
                        msg += f"\n{e.stderr}"
                    if e.stdout:
                        msg += f"\n{e.stdout}"
                raise click.ClickException(msg) from None
            except pydantic.ValidationError as e:
                raise click.ClickException(str(e)) from None
            except psycopg.OperationalError as e:
                logger.debug("an operational error occurred", exc_info=True)
                raise click.ClickException(str(e).strip()) from None


class Group(click.Group):
    command_class = Command
    group_class = type

    def add_command(self, cmd: click.Command, name: str | None = None) -> None:
        name = name or cmd.name
        assert name not in self.commands, f"command {name!r} already registered"
        super().add_command(cmd, name)

    def invoke(self, ctx: click.Context) -> Any:
        obj = ctx.ensure_object(Obj)
        settings = obj.ctx.settings
        pm = plugin_manager(settings)
        if not _install.check(pm, settings):
            raise click.ClickException(
                "broken installation; did you run 'site-configure install'?",
            )
        return super().invoke(ctx)


class PluggableCommandGroup(abc.ABC, Group):
    _plugin_commands_loaded: bool = False

    @abc.abstractmethod
    def register_plugin_commands(self, obj: Obj) -> None:
        ...

    def _load_plugins_commands(self, context: click.Context) -> None:
        if self._plugin_commands_loaded:
            return
        obj: Obj | None = context.obj
        if obj is None:
            try:
                obj = context.ensure_object(Obj)
            except InvalidSettingsError:
                return
        if obj is None:
            return
        self.register_plugin_commands(obj)
        self._plugin_commands_loaded = True

    def list_commands(self, context: click.Context) -> list[str]:
        self._load_plugins_commands(context)
        return super().list_commands(context)

    def get_command(self, context: click.Context, name: str) -> click.Command | None:
        self._load_plugins_commands(context)
        return super().get_command(context, name)


Callback = Callable[..., Any]

T = TypeVar("T", bound=CompositeManifest)
CommandFactory = Callable[[type[T]], Callback]


class CompositeCommandGroup(abc.ABC, Generic[T], Group):
    @property
    @abc.abstractmethod
    def model(self) -> type[T]:
        ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._composite_commands: dict[str, CommandFactory[T]] = {}
        super().__init__(*args, **kwargs)

    def command_with_composite_model(
        self, name: str
    ) -> Callable[[CommandFactory[T]], None]:
        """Decorator for callback that needs a composite model"""
        assert name not in self._composite_commands, name

        def decorator(factory: CommandFactory[T]) -> None:
            self._composite_commands[name] = factory

        return decorator

    def list_commands(self, context: click.Context) -> list[str]:
        return sorted(list(self._composite_commands) + super().list_commands(context))

    def get_command(self, context: click.Context, name: str) -> click.Command | None:
        try:
            factory = self._composite_commands[name]
        except KeyError:
            return super().get_command(context, name)
        else:
            if context.obj is None:  # During shell complete.
                model = self.model
            else:
                obj: Obj = context.obj
                p = plugin_manager(obj.ctx.settings)
                model = self.model.composite(p)
            f = factory(model)
            return click.command(cls=Command)(f)



from __future__ import annotations

import datetime
import json
import os
import pathlib
import socket
import subprocess
from collections.abc import Iterator
from typing import Callable

import psycopg
import pytest
import yaml
from psycopg.rows import dict_row

from .. import CertFactory


@pytest.fixture
def site_settings(
    tmp_path: pathlib.Path,
    playdir: pathlib.Path,
    pgbackrest_execpath: pathlib.Path | None,
    prometheus_execpath: pathlib.Path | None,
    ca_cert: pathlib.Path,
    cert_factory: CertFactory,
) -> Iterator[pathlib.Path]:
    with (playdir / "tutorial-settings.yaml.j2").open() as f:
        settings = yaml.safe_load(f)
    settings |= {
        "prefix": str(tmp_path),
        "run_prefix": str(tmp_path / "run"),
    }
    settings["postgresql"]["auth"]["passfile"] = str(tmp_path / "pgpass")
    if pgbackrest_execpath is not None:
        settings["pgbackrest"] |= {"execpath": str(pgbackrest_execpath)}
        settings["pgbackrest"]["repository"]["path"] = str(tmp_path / "pgbackups")
    else:
        pytest.skip("pgbackrest not available")
    if prometheus_execpath:
        settings["prometheus"] |= {"execpath": str(prometheus_execpath)}
    else:
        pytest.skip("prometheus not available")

    settings_f = tmp_path / "config.json"
    with settings_f.open("w") as f:
        json.dump(settings, f)

    env = os.environ.copy()
    env["SETTINGS"] = f"@{settings_f}"
    subprocess.run(
        ["pglift", "--debug", "site-configure", "install"],
        capture_output=True,
        check=True,
        env=env,
    )
    yield settings_f
    subprocess.run(
        ["pglift", "--non-interactive", "--debug", "site-configure", "uninstall"],
        capture_output=True,
        check=True,
        env=env,
    )


@pytest.fixture
def call_playbook(
    tmp_path: pathlib.Path,
    ansible_env: dict[str, str],
    ansible_vault: Callable[[dict[str, str]], pathlib.Path],
    playdir: pathlib.Path,
    site_settings: pathlib.Path,
) -> Iterator[Callable[[str], None]]:
    env = ansible_env.copy()

    env["SETTINGS"] = f"@{site_settings}"

    vault = ansible_vault(
        {
            "postgresql_surole_password": "supers3kret",
            "prod_bob_password": "s3kret",
            "backup_role_password": "b4ckup",
            "prometheus_role_password": "pr0m3th3u$",
        }
    )

    def call(playname: str) -> None:
        subprocess.check_call(
            [
                "ansible-playbook",
                "--extra-vars",
                f"@{vault}",
                playdir / playname,
            ],
            env=env,
        )

    yield call

    call("instances-delete.yml")


def cluster_name(dsn: str) -> str:
    with psycopg.connect(dsn, row_factory=dict_row) as cnx:
        cur = cnx.execute("SELECT setting FROM pg_settings WHERE name = 'cluster_name'")
        row = cur.fetchone()
        assert row
        name = row["setting"]
        assert isinstance(name, str), name
        return name


def test(call_playbook: Callable[[str], None]) -> None:
    call_playbook("instances-create.yml")

    prod_dsn = "host=/tmp user=postgres password=supers3kret dbname=postgres port=5433"
    assert cluster_name(prod_dsn) == "prod"
    with psycopg.connect(prod_dsn, row_factory=dict_row) as cnx:
        cnx.execute("SET TIME ZONE 'UTC'")
        cur = cnx.execute(
            "SELECT rolname,rolinherit,rolcanlogin,rolconnlimit,rolpassword,rolvaliduntil FROM pg_roles WHERE rolname = 'bob'"
        )
        assert cur.fetchone() == {
            "rolname": "bob",
            "rolinherit": True,
            "rolcanlogin": True,
            "rolconnlimit": 10,
            "rolpassword": "********",
            "rolvaliduntil": datetime.datetime(
                2025, 1, 1, tzinfo=datetime.timezone.utc
            ),
        }
        cur = cnx.execute(
            "SELECT r.rolname AS role, ARRAY_AGG(m.rolname ORDER BY m.rolname) AS member_of FROM pg_auth_members JOIN pg_authid m ON pg_auth_members.roleid = m.oid JOIN pg_authid r ON pg_auth_members.member = r.oid GROUP BY r.rolname"
        )
        assert cur.fetchall() == [
            {"role": "bob", "member_of": ["pg_read_all_stats", "pg_signal_backend"]},
            {"role": "peter", "member_of": ["pg_signal_backend"]},
            {
                "role": "pg_monitor",
                "member_of": [
                    "pg_read_all_settings",
                    "pg_read_all_stats",
                    "pg_stat_scan_tables",
                ],
            },
            {"role": "prometheus", "member_of": ["pg_monitor"]},
            {"role": "replication", "member_of": ["pg_read_all_stats"]},
        ]
        libraries = cnx.execute(
            "SELECT setting FROM pg_settings WHERE name = 'shared_preload_libraries';"
        ).fetchone()

    assert libraries is not None
    assert "pg_stat_statements" in libraries["setting"]

    socket.create_connection(("127.0.0.1", 9186), 1).close()

    # test connection with bob to the db database
    with psycopg.connect(
        "host=/tmp user=bob password=s3kret dbname=db port=5433", row_factory=dict_row
    ) as cnx:
        row = cnx.execute("SHOW work_mem").fetchone()
        extensions = cnx.execute("SELECT extname FROM pg_extension").fetchall()
    assert row is not None
    assert row["work_mem"] == "3MB"
    installed = [r["extname"] for r in extensions]
    assert "unaccent" in installed

    # check preprod cluster & postgres_exporter
    preprod_dsn = "host=/tmp user=postgres password=supers3kret dbname=test port=5434"
    assert cluster_name(preprod_dsn) == "preprod"
    socket.create_connection(("127.0.0.1", 9188), 1).close()

    # check dev cluster and postgres_exporter are stopped
    with pytest.raises(psycopg.OperationalError, match="No such file or directory"):
        cluster_name(
            "host=/tmp user=postgres password=supers3kret dbname=postgres port=5444"
        )
    with pytest.raises(ConnectionRefusedError):
        socket.create_connection(("127.0.0.1", 9189), 1)

    call_playbook("instances-update.yml")

    # prod running
    assert cluster_name(prod_dsn) == "prod"

    # pg_stat_statements extension is uninstalled
    with psycopg.connect(prod_dsn, row_factory=dict_row) as cnx:
        libraries = cnx.execute(
            "SELECT setting FROM pg_settings WHERE name = 'shared_preload_libraries';"
        ).fetchone()
    assert libraries is not None
    assert "pg_stat_statements" not in libraries["setting"]

    # bob user and db database no longer exists
    with pytest.raises(
        psycopg.OperationalError, match='password authentication failed for user "bob"'
    ):
        with psycopg.connect(
            "host=/tmp user=bob password=s3kret dbname=template1 port=5433"
        ):
            pass
    with pytest.raises(psycopg.OperationalError, match='database "db" does not exist'):
        with psycopg.connect(
            "host=/tmp user=postgres password=supers3kret dbname=db port=5433"
        ):
            pass
    with pytest.raises(psycopg.OperationalError, match='database "db2" does not exist'):
        with psycopg.connect(
            "host=/tmp user=postgres password=supers3kret dbname=db2 port=5433"
        ):
            pass
    with psycopg.connect(prod_dsn, row_factory=dict_row) as cnx:
        row = cnx.execute(
            "SELECT pg_catalog.pg_get_userbyid(d.datdba) as owner FROM pg_catalog.pg_database d WHERE d.datname = 'db3'"
        ).fetchone()
    assert row is not None
    assert row["owner"] == "postgres"

    # preprod stopped
    with pytest.raises(psycopg.OperationalError, match="No such file or directory"):
        assert cluster_name(preprod_dsn) == "preprod"
    with pytest.raises(ConnectionRefusedError):
        socket.create_connection(("127.0.0.1", 9188), 1)

    # dev running
    dev_dsn = "host=/tmp user=postgres password=supers3kret dbname=postgres port=5455"
    assert cluster_name(dev_dsn) == "dev"
    socket.create_connection(("127.0.0.1", 9189)).close()



from __future__ import annotations

import asyncio
import contextlib
import logging
import subprocess
import sys
from pathlib import Path
from unittest.mock import patch

import pytest

from pglift import cmd
from pglift.exceptions import CommandError, SystemError
from pglift.types import Status


def test_execute_program(caplog: pytest.LogCaptureFixture, tmp_path: Path) -> None:
    command = ["/c", "m", "d"]
    with patch("os.execve", autospec=True) as execve, patch(
        "os.execv", autospec=True
    ) as execv:
        cmd.execute_program(command, env={"X": "Y"})
        execve.assert_called_once_with("/c", command, {"X": "Y"})
        assert not execv.called
    with patch("os.execve", autospec=True) as execve, patch(
        "os.execv", autospec=True
    ) as execv, caplog.at_level(logging.DEBUG, logger="pglift.cmd"):
        cmd.execute_program(command)
        execv.assert_called_once_with("/c", command)
        assert not execve.called
    assert "executing program '/c m d'" in caplog.records[0].message


def test_start_program_terminate_program_status_program(
    caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    pidfile = tmp_path / "sleep" / "pid"
    p = cmd.Program(["sleep", "10"], pidfile, timeout=0.01, env={"X_DEBUG": "1"})
    assert p.pidfile == pidfile
    with pidfile.open() as f:
        pid = f.read()

    assert p.proc.pid == int(pid)

    proc = Path("/proc") / pid
    assert proc.exists()
    assert "sleep\x0010\x00" in (proc / "cmdline").read_text()
    assert "X_DEBUG" in (proc / "environ").read_text()

    assert cmd.status_program(pidfile) == Status.running

    with pidfile.open("a") as f:
        f.write("\nextra\ninformation\nignored")
    assert cmd.status_program(pidfile) == Status.running

    with pytest.raises(SystemError, match="running already"):
        cmd.Program(["sleep", "10"], pidfile)

    cmd.terminate_program(pidfile)
    r = subprocess.run(["pgrep", pid], check=False)
    assert r.returncode == 1

    assert not pidfile.exists()
    assert cmd.status_program(pidfile) == Status.not_running

    pidfile = tmp_path / "invalid.pid"
    pidfile.write_text("innnnvaaaaaaaaaaliiiiiiiiiiid")
    assert cmd.status_program(pidfile) == Status.not_running
    caplog.clear()
    with pytest.raises(CommandError) as excinfo, caplog.at_level(
        logging.DEBUG, logger=__name__
    ):
        cmd.Program(["sleep", "well"], pidfile, env={"LANG": "C", "LC_ALL": "C"})
    assert not pidfile.exists()
    assert "sleep is supposed to be running" in caplog.records[0].message
    assert "sleep: invalid time interval 'well'" in caplog.records[2].message
    assert "sleep: invalid time interval 'well'" in excinfo.value.stderr

    pidfile = tmp_path / "notfound"
    caplog.clear()
    with caplog.at_level(logging.WARNING, logger="pglift.cmd"):
        cmd.terminate_program(pidfile)
    assert f"program from {pidfile} not running" in caplog.records[0].message


def test_program_context(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    pidfile = tmp_path / "pid"
    pyprog = tmp_path / "prog.py"
    pyprog.write_text(
        "\n".join(
            [
                "import logging, time, signal, sys",
                "signal.signal(",
                "   signal.SIGTERM,",
                "   lambda signum, frame: logging.error('got signal %d', signum),",
                ")",
                "s = float(sys.argv[1])",
                "logging.warning('sleeping %.1fs', s)",
                "time.sleep(s)",
            ]
        )
    )
    with pytest.raises(ValueError, match="expected"), caplog.at_level(
        logging.DEBUG, logger="pglift.cmd"
    ):
        with cmd.Program(
            [sys.executable, str(pyprog), "0.2"], pidfile, timeout=0.1
        ) as prog:
            assert pidfile.exists()
            assert prog.proc.returncode is None
            raise ValueError("expected")
    messages = caplog.messages
    startuplog, pylog, termlog, signallog = messages
    assert startuplog.startswith("starting program '")
    assert "sleeping 0.2s" in pylog
    assert "terminating program" in termlog
    assert "got signal 15" in signallog
    assert not pidfile.exists()

    class SomeError(Exception):
        pass

    loop = asyncio.get_event_loop()
    program = sys.executable
    with contextlib.suppress(SomeError), caplog.at_level(
        logging.DEBUG, logger="pglift.cmd"
    ):
        with cmd.Program([program, "--version"], pidfile=None, timeout=0) as prog:
            loop.run_until_complete(prog.proc.wait())
            raise SomeError
    terminating_msg, terminated_msg = caplog.messages[-2:]
    assert terminating_msg == f"terminating program '{program} --version'"
    assert terminated_msg == f"program {program} already terminated"



from __future__ import annotations

import pytest

from pglift import exceptions, roles
from pglift.ctx import Context
from pglift.models import interface
from pglift.models.system import Instance


def test_standby_role_drop(ctx: Context, standby_instance: Instance) -> None:
    role = interface.Role(name="alice")
    with pytest.raises(
        exceptions.InstanceReadOnlyError,
        match=f"^{standby_instance.version}/standby is a read-only standby instance$",
    ):
        roles.drop(ctx, standby_instance, role)



from __future__ import annotations

import logging
from pathlib import Path

from pgtoolkit.conf import Configuration

from pglift import util

from .. import hookimpl
from ..models import system
from ..pm import PluginManager
from ..settings import Settings, _rsyslog

logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    return settings.rsyslog is not None


def get_settings(settings: Settings) -> _rsyslog.Settings:
    assert settings.rsyslog is not None
    return settings.rsyslog


def config_path(settings: _rsyslog.Settings) -> Path:
    return settings.configdir / "rsyslog.conf"


@hookimpl
def site_configure_install(settings: Settings, pm: PluginManager) -> None:
    logger.info("creating rsyslog config directory")
    s = get_settings(settings)
    s.configdir.mkdir(mode=0o750, exist_ok=True, parents=True)
    results = pm.hook.rsyslog_config(settings=settings)
    with config_path(s).open("w") as f:
        logger.info("writing rsyslog config")
        f.write("\n".join(results))


@hookimpl
def site_configure_uninstall(settings: Settings) -> None:
    logger.info("deleting rsyslog config directory")
    s = get_settings(settings)
    util.rmtree(s.configdir)


@hookimpl
def site_configure_installed(settings: Settings) -> bool:
    s = get_settings(settings)
    if not (fpath := config_path(s)).exists():
        logger.error("rsyslog configuration file '%s' missing", fpath)
        return False
    return True


@hookimpl
def instance_settings(instance: system.BaseInstance) -> Configuration:
    pgconfig = util.template("postgresql", "postgresql-rsyslog.conf").format(
        name=instance.name,
        version=instance.version,
    )
    config = Configuration()
    list(config.parse(pgconfig.splitlines()))
    return config



from __future__ import annotations

from ..settings import Settings, _powa

POWA_EXTENSIONS = [
    "btree_gist",
    "pg_qualstats",
    "pg_stat_statements",
    "pg_stat_kcache",
    "powa",
]

# Libraries to load in shared_preload_libraries and to install.
# Order is important here, for example `pg_stat_statements` needs to be loaded
# before `pg_stat_kcache` in shared_preload_libraries
POWA_LIBRARIES = ["pg_qualstats", "pg_stat_statements", "pg_stat_kcache"]


def available(settings: Settings) -> _powa.Settings | None:
    return settings.powa



from __future__ import annotations

import configparser
import logging
from pathlib import Path

import click
from pgtoolkit import conf as pgconf

from .. import cmd, exceptions, hookimpl, postgresql, types, util
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings, _pgbackrest
from ..task import task
from ..types import BackupType, CompletedProcess
from . import base, models
from . import register_if as base_register_if
from . import role
from .base import get_settings

PathRepository = _pgbackrest.PathRepository
logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    if not base_register_if(settings):
        return False
    s = get_settings(settings)
    return isinstance(s.repository, PathRepository)


@hookimpl
def site_configure_install(settings: Settings) -> None:
    s = get_settings(settings)
    base.site_configure_install(settings, base_config(s))
    logger.info("creating pgbackrest repository path")
    repository_settings(s).path.mkdir(exist_ok=True, parents=True)


@hookimpl
def site_configure_uninstall(settings: Settings, ctx: Context) -> None:
    s = get_settings(settings)
    base.site_configure_uninstall(settings)
    util.rmdir(s.configpath)
    rs = repository_settings(s)
    if ctx.confirm(f"Delete pgbackrest repository path {rs.path}?", False):
        logger.info("deleting pgbackrest repository path")
        util.rmtree(rs.path)


@hookimpl
def site_configure_uninstalled(settings: Settings) -> None:
    s = get_settings(settings)
    rs = repository_settings(s)
    for f in (base.config_directory(s), base.base_configpath(s), rs.path):
        if not f.exists():
            raise exceptions.InstallationError(
                f"pgBackRest configuration path {f} missing"
            )


@hookimpl
def instance_configured(
    ctx: Context,
    manifest: interface.Instance,
    config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    upgrading_from: system.Instance | None,
) -> None:
    with base.instance_configured(
        ctx, manifest, config, changes, creating, upgrading_from
    ) as instance:
        settings = get_settings(instance._settings)
        service_manifest = manifest.service_manifest(models.ServiceManifest)
        service = base.service(instance, service_manifest, settings, upgrading_from)
        if creating and upgrading_from is None and base.enabled(instance, settings):
            if not ctx.confirm(
                f"Stanza {service.stanza!r} already bound to another instance, continue by overwriting it?",
                False,
            ):
                raise exceptions.Cancelled("pgbackrest repository already exists")
            revert_init(service, settings, instance.datadir)
            base.revert_setup(
                ctx, service, settings, config, {}, False, instance.datadir
            )

    if upgrading_from is not None:
        upgrade(service, settings)
    elif creating:
        init(service, settings, instance.datadir)

    if creating and postgresql.is_running(instance):
        password = None
        backup_role = role(instance._settings, manifest)
        assert backup_role is not None
        if backup_role.password is not None:
            password = backup_role.password.get_secret_value()
        if instance.standby:
            logger.warning("not checking pgBackRest configuration on a standby")
        else:
            base.check(instance, service, settings, password)


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    with base.instance_dropped(ctx, instance) as service:
        if not service:
            return
        settings = get_settings(instance._settings)
        if not (nb_backups := len(base.backup_info(service, settings)["backup"])) or (
            can_delete_stanza(service, settings, instance.datadir)
            and ctx.confirm(
                f"Confirm deletion of {nb_backups} backup(s) for stanza {service.stanza}?",
                False,
            )
        ):
            delete_stanza(service, settings, instance.datadir)


@hookimpl
def instance_cli(group: click.Group) -> None:
    from .cli import instance_backup

    group.add_command(instance_backup)


def repository_settings(settings: _pgbackrest.Settings) -> PathRepository:
    assert isinstance(settings.repository, PathRepository)
    return settings.repository


def base_config(settings: _pgbackrest.Settings) -> configparser.ConfigParser:
    cp = base.parser()
    cp.read_string(
        util.template("pgbackrest", "pgbackrest.conf").format(**dict(settings))
    )
    s = repository_settings(settings)
    cp["global"]["repo1-path"] = str(s.path)
    for opt, value in s.retention:
        cp["global"][f"repo1-retention-{opt}"] = str(value)
    return cp


@task(title="creating pgBackRest stanza {service.stanza}")
def init(
    service: models.Service,
    settings: _pgbackrest.Settings,
    datadir: Path,
) -> None:
    cmd.run(
        base.make_cmd(service.stanza, settings, "stanza-create", "--no-online"),
        check=True,
    )


@init.revert(title="deleting pgBackRest stanza {service.stanza}")
def revert_init(
    service: models.Service,
    settings: _pgbackrest.Settings,
    datadir: Path,
) -> None:
    if not can_delete_stanza(service, settings, datadir):
        logger.debug(
            "not deleting stanza %s, still used by another instance", service.stanza
        )
        return
    delete_stanza(service, settings, datadir)


def can_delete_stanza(
    service: models.Service, settings: _pgbackrest.Settings, datadir: Path
) -> bool:
    for idx, path in base.stanza_pgpaths(service.path, service.stanza):
        if (idx, path) != (service.index, datadir):
            return False
    return True


def delete_stanza(
    service: models.Service, settings: _pgbackrest.Settings, datadir: Path
) -> None:
    stanza = service.stanza
    cmd.run(base.make_cmd(stanza, settings, "stop"), check=True)
    cmd.run(
        base.make_cmd(
            stanza, settings, "stanza-delete", "--pg1-path", str(datadir), "--force"
        ),
        check=True,
    )


def upgrade(service: models.Service, settings: _pgbackrest.Settings) -> None:
    """Upgrade stanza"""
    stanza = service.stanza
    logger.info("upgrading pgBackRest stanza %s", stanza)
    cmd.run(
        base.make_cmd(stanza, settings, "stanza-upgrade", "--no-online"), check=True
    )


def backup_command(
    instance: system.Instance,
    settings: _pgbackrest.Settings,
    *,
    type: BackupType = BackupType.default(),  # noqa: B008
    start_fast: bool = True,
    backup_standby: bool = False,
) -> list[str]:
    """Return the full pgbackrest command to perform a backup for ``instance``.

    :param type: backup type (one of 'full', 'incr', 'diff').

    Ref.: https://pgbackrest.org/command.html#command-backup
    """
    args = [f"--type={type.name}", "backup"]
    if start_fast:
        args.insert(-1, "--start-fast")
    if backup_standby:
        args.insert(-1, "--backup-standby")
    s = instance.service(models.Service)
    return base.make_cmd(s.stanza, settings, *args)


def backup(
    instance: system.Instance,
    settings: _pgbackrest.Settings,
    *,
    type: BackupType = BackupType.default(),  # noqa: B008
) -> CompletedProcess:
    """Perform a backup of ``instance``.

    :param type: backup type (one of 'full', 'incr', 'diff').

    Ref.: https://pgbackrest.org/command.html#command-backup
    """
    logger.info("backing up instance %s with pgBackRest", instance)
    cmd_args = backup_command(
        instance, settings, type=type, backup_standby=instance.standby is not None
    )
    postgresql_settings = instance._settings.postgresql
    env = postgresql.ctl.libpq_environ(instance, postgresql_settings.backuprole.name)
    return cmd.run(cmd_args, check=True, env=env)



from __future__ import annotations

from collections.abc import Sequence

import psycopg.rows
from psycopg import sql

from . import db
from .ctx import Context
from .models import system


def inspect_privileges(
    ctx: Context,
    instance: system.Instance,
    database: str,
    roles: Sequence[str] = (),
    defaults: bool = False,
) -> list[system.DefaultPrivilege] | list[system.Privilege]:
    args = {}
    where_clause = sql.SQL("")
    if roles:
        where_clause = sql.SQL("AND pg_roles.rolname = ANY(%(roles)s)")
        args["roles"] = list(roles)
    rtype: type[system.DefaultPrivilege] = (
        system.DefaultPrivilege if defaults else system.Privilege
    )
    with db.connect(instance, ctx=ctx, dbname=database) as cnx:
        with cnx.cursor(row_factory=psycopg.rows.class_row(rtype)) as cur:
            cur.execute(db.query(rtype.query, where_clause=where_clause), args)
            return cur.fetchall()


def get(
    ctx: Context,
    instance: system.Instance,
    *,
    databases: Sequence[str] = (),
    roles: Sequence[str] = (),
    defaults: bool = False,
) -> list[system.DefaultPrivilege] | list[system.Privilege]:
    """List access privileges for databases of an instance.

    :param databases: list of databases to inspect (all will be inspected if
        unspecified).
    :param roles: list of roles to restrict inspection on.
    :param defaults: if ``True``, get default privileges.

    :raises ValueError: if an element of `databases` or `roles` does not
        exist.
    """

    with db.connect(instance, ctx=ctx) as cnx:
        cur = cnx.execute(db.query("database_list", where_clause=sql.SQL("")))
        existing_databases = [db["name"] for db in cur.fetchall()]
    if not databases:
        databases = existing_databases
    else:
        if unknown_dbs := set(databases) - set(existing_databases):
            raise ValueError(f"database(s) not found: {', '.join(unknown_dbs)}")

    if roles:
        with db.connect(instance, ctx=ctx) as cnx:
            cur = cnx.execute(db.query("role_list_names"))
            existing_roles = [n["rolname"] for n in cur.fetchall()]
        if unknown_roles := set(roles) - set(existing_roles):
            raise ValueError(f"role(s) not found: {', '.join(unknown_roles)}")

    return [
        prvlg
        for database in databases
        for prvlg in inspect_privileges(
            ctx, instance, database, roles=roles, defaults=defaults
        )
    ]



from __future__ import annotations

import logging
from collections.abc import Iterator
from typing import Any

from . import execpath, hookimpl, systemd
from .ctx import Context
from .models import interface, system
from .pgbackrest import repo_path
from .settings import Settings

logger = logging.getLogger(__name__)
service_name = "backup"
BACKUP_SERVICE_NAME = "pglift-backup@.service"
BACKUP_TIMER_NAME = "pglift-backup@.timer"


def register_if(settings: Settings) -> bool:
    return (
        settings.service_manager == "systemd"
        and settings.scheduler == "systemd"
        and settings.pgbackrest is not None
        and repo_path.register_if(settings)
    )


@hookimpl
def systemd_unit_templates(
    settings: Settings, env: dict[str, Any], content: bool
) -> Iterator[str] | Iterator[tuple[str, str]]:
    if not content:
        yield BACKUP_SERVICE_NAME
        yield BACKUP_TIMER_NAME
        return

    yield BACKUP_SERVICE_NAME, systemd.template(BACKUP_SERVICE_NAME).format(
        executeas=systemd.executeas(settings),
        environment=systemd.environment(env),
        execpath=execpath,
    )
    yield BACKUP_TIMER_NAME, systemd.template(BACKUP_TIMER_NAME)


@hookimpl
def instance_configured(ctx: Context, manifest: interface.Instance) -> None:
    """Enable scheduled backup job for configured instance."""
    instance = system.Instance.system_lookup(
        (manifest.name, manifest.version, ctx.settings)
    )
    ctx.hook.schedule_service(ctx=ctx, service=service_name, name=instance.qualname)


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    """Disable scheduled backup job when instance is being dropped."""
    ctx.hook.unschedule_service(
        ctx=ctx, service=service_name, name=instance.qualname, now=True
    )


@hookimpl
def instance_started(ctx: Context, instance: system.Instance) -> None:
    """Start schedule backup job at instance startup."""
    ctx.hook.start_timer(ctx=ctx, service=service_name, name=instance.qualname)


@hookimpl
def instance_stopped(ctx: Context, instance: system.Instance) -> None:
    """Stop schedule backup job when instance is stopping."""
    ctx.hook.stop_timer(ctx=ctx, service=service_name, name=instance.qualname)



from __future__ import annotations

import logging
import re

import pytest

from pglift import exceptions, task


class SimpleDisplayer:
    def __init__(self) -> None:
        self.records: list[str] = []

    def clear(self) -> None:
        self.records.clear()

    def handle(self, msg: str) -> None:
        self.records.append(msg)


def test_task() -> None:
    @task.task(title="negate")
    def neg(x: int) -> int:
        return -x

    assert re.match(r"<Task 'neg' at 0x(\d+)>" "", repr(neg))

    assert neg(1) == -1
    assert neg.revert_action is None

    # Check static and runtime type error.
    with pytest.raises(TypeError):
        neg("1")  # type: ignore[arg-type]

    @neg.revert(title="negate again")
    def revert_neg(x: int) -> int:
        return -x

    assert neg.revert_action
    assert neg.revert_action.call(-1) == 1


def test_display() -> None:
    ...


def test_transaction_state() -> None:
    with pytest.raises(RuntimeError, match="inconsistent task state"):
        with task.transaction():
            with task.transaction():
                pass

    with pytest.raises(ValueError, match="expected"):
        with task.transaction():
            assert task.Task._calls is not None
            raise ValueError("expected")
    assert task.Task._calls is None


def test_transaction(caplog: pytest.LogCaptureFixture) -> None:
    values = set()

    @task.task(title="add {x} to values")
    def add(x: int, fail: bool = False) -> None:
        values.add(x)
        if fail:
            raise RuntimeError("oups")

    add(1)
    assert values == {1}

    displayer = SimpleDisplayer()
    with pytest.raises(RuntimeError, match="oups"):
        with task.displayer_installed(displayer), task.transaction():
            add(2, fail=True)
    # no revert action
    assert values == {1, 2}
    assert displayer.records == ["add 2 to values"]

    displayer.clear()

    @add.revert(title="remove {x} from values (fail={fail})")
    def remove(x: int, fail: bool = False) -> None:
        try:
            values.remove(x)
        except KeyError:
            pass

    with pytest.raises(RuntimeError, match="oups"):
        with task.displayer_installed(displayer), task.transaction():
            add(3, fail=False)
            add(4, fail=True)
    assert values == {1, 2}
    assert displayer.records == [
        "add 3 to values",
        "add 4 to values",
        "remove 4 from values (fail=True)",
        "remove 3 from values (fail=False)",
    ]

    displayer.clear()

    @add.revert(title="remove numbers, failed")
    def remove_fail(x: int, fail: bool = False) -> None:
        values.remove(x)
        if fail:
            raise ValueError("failed to fail")

    caplog.clear()
    with pytest.raises(ValueError, match="failed to fail"), caplog.at_level(
        logging.WARNING
    ):
        with task.transaction():
            add(3, fail=False)
            add(4, fail=True)
    assert values == {1, 2, 3}
    assert caplog.messages == [
        "oups",
        "reverting: add 4 to values",
    ]

    with pytest.raises(RuntimeError, match="oups"):
        with task.transaction(False):
            add(4, fail=True)
    assert values == {1, 2, 3, 4}

    @task.task
    def intr() -> None:
        raise KeyboardInterrupt

    caplog.clear()
    with pytest.raises(KeyboardInterrupt), caplog.at_level(logging.WARNING):
        with task.transaction():
            intr()
    assert caplog.messages == [f"{intr} interrupted"]

    @task.task
    def cancel() -> None:
        raise exceptions.Cancelled("forget about it")

    caplog.clear()
    with pytest.raises(exceptions.Cancelled):
        with task.transaction():
            cancel()
    assert not caplog.messages


def test_check_task_revert_signature() -> None:
    @task.task
    def t(x: int) -> None:
        pass

    with pytest.raises(
        AssertionError,
        match=(
            r"Parameters of function tests.unit.test_task.t\(\(x: 'int'\) -> 'None'\) "
            r"differ from related revert function tests.unit.test_task.rt\(\(x: 'str'\) -> 'None'\)"
        ),
    ):

        @t.revert  # type: ignore[arg-type]
        def rt(x: str) -> None:
            pass
